# -*- coding: utf-8 -*-
_tdl_no_reimport = True

import os.path
import time

import Kittens.utils
import Kittens.widgets
from PyQt4.Qt import (QWidget, QDialog, QWizard, QWizardPage, QButtonGroup, QVBoxLayout,
                      QRadioButton, QObject, SIGNAL, QHBoxLayout, QLineEdit, QPushButton,
                      QFileDialog, QMessageBox, QHeaderView, QAbstractItemView,
                      QFontMetrics, QFont, QTreeWidget, QSizePolicy, QMenu, QMimeData, QUrl,
                      QComboBox, QMimeData, QTreeWidgetItem, Qt, QApplication,
                      QClipboard, QLabel, QSplitter, QTextEdit, QTextDocument, QSize,
                      QFrame, QStackedWidget, QWidgetAction, QMenu, QTextBrowser, QPoint,
                      QDrag, QListWidget, QListWidgetItem, QMainWindow, QToolBar, QTimer,
                      QCoreApplication, QEventLoop, QCursor, QListWidget)
import six
if six.PY3:
    QVariant = str
else:
    from PyQt4.Qt import (QVariant)

import Purr
import Purr.Editors
import Purr.LogEntry
import Purr.Pipe
import Purr.RenderIndex
from Purr import Config, pixmaps, dprint


class BusyIndicator(object):
    def __init__(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

    def __del__(self):
        QApplication.restoreOverrideCursor()


class HTMLViewerDialog(QDialog):
    """This class implements a dialog to view a piece of HTML text."""

    def __init__(self, parent, config_name=None, buttons=[], *args):
        """Creates dialog.
        'config_name' is used to get/set default window size from Config object
        'buttons' can be a list of names or (QPixmapWrapper,name[,tooltip]) tuples to provide
        custom buttons at the bottom of the dialog. When a button is clicked, the dialog
        emits SIGNAL("name").
        A "Close" button is always provided, this simply hides the dialog.
        """
        QDialog.__init__(self, parent, *args)
        self.setModal(False)
        lo = QVBoxLayout(self)
        # create viewer
        self.label = QLabel(self)
        self.label.setMargin(5)
        self.label.setWordWrap(True)
        lo.addWidget(self.label)
        self.label.hide()
        self.viewer = QTextBrowser(self)
        lo.addWidget(self.viewer)
        # self.viewer.setReadOnly(True)
        self.viewer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        QObject.connect(self.viewer, SIGNAL("anchorClicked(const QUrl &)"), self._urlClicked)
        self._source = None
        lo.addSpacing(5)
        # create button bar
        btnfr = QFrame(self)
        btnfr.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        # btnfr.setMargin(5)
        lo.addWidget(btnfr)
        lo.addSpacing(5)
        btnfr_lo = QHBoxLayout(btnfr)
        btnfr_lo.setMargin(5)
        # add user buttons
        self._user_buttons = {}
        for name in buttons:
            if isinstance(name, str):
                btn = QPushButton(name, btnfr)
            elif isinstance(name, (list, tuple)):
                if len(name) < 3:
                    pixmap, name = name
                    tip = None
                else:
                    pixmap, name, tip = name
                btn = QPushButton(pixmap.icon(), name, btnfr)
                if tip:
                    btn.setToolTip(tip)
            self._user_buttons[name] = btn
            btn._clicked = Kittens.utils.curry(self.emit, SIGNAL(name))
            self.connect(btn, SIGNAL("clicked()"), btn._clicked)
            btnfr_lo.addWidget(btn, 1)
        # add a Close button
        btnfr_lo.addStretch(100)
        closebtn = QPushButton(pixmaps.grey_round_cross.icon(), "Close", btnfr)
        self.connect(closebtn, SIGNAL("clicked()"), self.hide)
        btnfr_lo.addWidget(closebtn, 1)
        # resize selves
        self.config_name = config_name or "html-viewer"
        width = Config.getint('%s-width' % self.config_name, 512)
        height = Config.getint('%s-height' % self.config_name, 512)
        self.resize(QSize(width, height))

    def resizeEvent(self, ev):
        QDialog.resizeEvent(self, ev)
        sz = ev.size()
        Config.set('%s-width' % self.config_name, sz.width())
        Config.set('%s-height' % self.config_name, sz.height())

    def setDocument(self, filename, empty=""):
        """Sets the HTML text to be displayed. """
        self._source = QUrl.fromLocalFile(filename)
        if os.path.exists(filename):
            self.viewer.setSource(self._source)
        else:
            self.viewer.setText(empty)

    def _urlClicked(self, url):
        path = str(url.path())
        if path:
            self.emit(SIGNAL("viewPath"), path)
        # to make sure it keeps displaying the same thing
        self.viewer.setSource(self._source)

    def reload(self):
        self.viewer.reload()

    def setLabel(self, label=None):
        if label is None:
            self.label.hide()
        else:
            self.label.setText(label)
            self.label.show()

            # self.watching = bool(enable)
            # if enable:
            #  self.purrer.enableWatching(self.path)
            # else:
            #  self.purrer.disableWatching(self.path)
            # self.mainwin._checkPounceStatus()


class DirectoryListWidget(Kittens.widgets.ClickableListWidget):
    """This class implements a QTreeWidget for data products.
    """

    def __init__(self, *args):
        Kittens.widgets.ClickableListWidget.__init__(self, *args)
        # insert columns, and numbers for them
        # setup other properties of the listview
        self.setAcceptDrops(True)
        self.setSelectionMode(QListWidget.SingleSelection)
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        QObject.connect(self, SIGNAL("itemSelectionChanged()"), self._selectionChanged)
        QObject.connect(self, SIGNAL("itemChanged(QListWidgetItem *)"), self._itemChanged)
        self._items = {}
        self._item_height = None
        self._max_height_items = 5
        self._min_height = 8
        self.setMaximumSize(1000000, self._min_height)

    class DirItem(QListWidgetItem):
        # these represent the watch-states
        ToCheckState = {Purr.UNWATCHED: Qt.Unchecked, Purr.WATCHED: Qt.PartiallyChecked, Purr.POUNCE: Qt.Checked}
        FromCheckState = dict([(state, watch) for watch, state in list(ToCheckState.items())])

        def __init__(self, pathname, parent=None, watching=Purr.WATCHED):
            self._pathname = pathname
            pathname = Kittens.utils.collapseuser(pathname)
            self._in_setWatching = True
            QListWidgetItem.__init__(self, pathname, parent)
            self.setFlags(Qt.ItemIsSelectable | Qt.ItemIsUserCheckable | Qt.ItemIsEnabled | Qt.ItemIsTristate)
            self.setWatching(watching)

        def pathname(self):
            return self._pathname

        def watching(self):
            return self.FromCheckState[self.checkState()]

        def setWatching(self, watching):
            self._in_setWatching = True
            self._watching = watching
            self.setCheckState(self.ToCheckState[watching])
            self._in_setWatching = False

    def _selectionChanged(self):
        sel = self.selectedItems()
        self.emit(SIGNAL("directorySelected"), sel and sel[0].pathname())
        self.emit(SIGNAL("hasSelection"), bool(sel))

    def _itemChanged(self, item):
        if not getattr(item, '_in_setWatching', False):
            item.setWatching((item._watching + 1) % (Purr.POUNCE + 1))
            self.emit(SIGNAL("directoryStateChanged"), item.pathname(), item.watching())

    def _checkSize(self):
        """Automatically resizes widget to display at most max_height_items items"""
        if self._item_height is not None:
            sz = min(self._max_height_items, self.count()) * self._item_height + 5
            sz = max(sz, 20)
            self.setMinimumSize(0, sz)
            self.setMaximumSize(1000000, sz)
            self.resize(self.width(), sz)

    def clear(self):
        Kittens.widgets.ClickableListWidget.clear(self)
        self._items = {}

    def add(self, pathname, watching=Purr.WATCHED):
        if pathname in self._items:
            return
        item = self._items[pathname] = self.DirItem(pathname, parent=self, watching=watching)
        # init item height, if inserting first item
        if self._item_height is None:
            self._item_height = self.visualItemRect(item).height()
        self._checkSize()

    def remove(self, pathname):
        item = self._items.get(pathname, None)
        return item and self._removeItem(item)

    def _removeItem(self, item):
        del self._items[item.pathname()]
        self.emit(SIGNAL("directoryStateChanged"), item.pathname(), Purr.REMOVED)
        self.takeItem(self.row(item))
        self._checkSize()

    def removeCurrent(self, confirm=True):
        sel = self.selectedItems()
        if sel:
            if confirm:
                if QMessageBox.warning(self, "Detaching from directory",
                                       """Do you really want to stop monitoring the directory <tt>%s</tt>?""" % sel[
                                           0].pathname(),
                                       QMessageBox.Yes, QMessageBox.No) == QMessageBox.Yes:
                    return self._removeItem(sel[0])
        return None


class LogEntryTree(Kittens.widgets.ClickableTreeWidget):
    def __init__(self, *args):
        Kittens.widgets.ClickableTreeWidget.__init__(self, *args)
        self.setDragEnabled(True)

    def mimeTypes(self):
        return ["text/x-url"]

    def mimeData(self, itemlist):
        mimedata = QMimeData()
        urls = []
        for item in itemlist:
            dp = getattr(item, "_dp", None)
            dp and urls.append(QUrl.fromLocalFile(dp.fullpath or dp.sourcepath))
        mimedata.setUrls(urls)
        return mimedata


class MainWindow(QMainWindow):
    about_message = """
    <P>PURR ("<B>P</B>URR is <B>U</B>seful for <B>R</B>emembering <B>R</B>eductions", for those working with
    a stable version, or "<B>P</B>URR <B>U</B>sually <B>R</B>emembers <B>R</B>eductions", for those
    working with a development version, or "<B>P</B>URR <B>U</B>sed to <B>R</B>emember <B>R</B>eductions",
    for those working with a broken version) is a tool for
    automatically keeping a log of your data reduction operations. PURR will monitor your working directories
    for new or updated files (called "data products"), and upon seeing any, it can "pounce" -- that is, offer
    you the option of saving the files to a log, along with descriptive comments. It will then
    generate an HTML page with a pretty rendering of your log and data products.</P>
  """

    def __init__(self, parent, hide_on_close=False):
        QMainWindow.__init__(self, parent)
        self._hide_on_close = hide_on_close
        # replace the BusyIndicator class with a GUI-aware one
        Purr.BusyIndicator = BusyIndicator
        self._pounce = False
        # we keep a small stack of previously active purrers. This makes directory changes
        # faster (when going back and forth between dirs)
        # current purrer
        self.purrer = None
        self.purrer_stack = []
        # Purr pipes for receiving remote commands
        self.purrpipes = {}
        # init GUI
        self.setWindowTitle("PURR")
        self.setWindowIcon(pixmaps.purr_logo.icon())
        cw = QWidget(self)
        self.setCentralWidget(cw)
        cwlo = QVBoxLayout(cw)
        cwlo.setContentsMargins(0, 0, 0, 0)
        cwlo.setMargin(5)
        cwlo.setSpacing(0)
        toplo = QHBoxLayout();
        cwlo.addLayout(toplo)

        # About dialog
        self._about_dialog = QMessageBox(self)
        self._about_dialog.setWindowTitle("About PURR")
        self._about_dialog.setText(self.about_message + """
        <P>PURR is not watching any directories right now. You may need to restart it, and give it
  some directory names on the command line.</P>""")
        self._about_dialog.setIconPixmap(pixmaps.purr_logo.pm())
        # Log viewer dialog
        self.viewer_dialog = HTMLViewerDialog(self, config_name="log-viewer",
                                              buttons=[(pixmaps.blue_round_reload, "Regenerate",
                                                        """<P>Regenerates your log's HTML code from scratch. This can be useful if
                                                        your PURR version has changed, or if there was an error of some kind
                                                        the last time the files were generated.</P>
                                                        """)])
        self._viewer_timestamp = None
        self.connect(self.viewer_dialog, SIGNAL("Regenerate"), self._regenerateLog)
        self.connect(self.viewer_dialog, SIGNAL("viewPath"), self._viewPath)

        # Log title toolbar
        title_tb = QToolBar(cw)
        title_tb.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        title_tb.setIconSize(QSize(16, 16))
        cwlo.addWidget(title_tb)
        title_label = QLabel("Purrlog title:", title_tb)
        title_tb.addWidget(title_label)
        self.title_editor = QLineEdit(title_tb)
        title_tb.addWidget(self.title_editor)
        self.connect(self.title_editor, SIGNAL("editingFinished()"), self._titleChanged)
        tip = """<P>This is your current log title. To rename the log, enter new name here and press Enter.</P>"""
        title_label.setToolTip(tip)
        self.title_editor.setToolTip(tip)
        self.wviewlog = title_tb.addAction(pixmaps.openbook.icon(), "View", self._showViewerDialog)
        self.wviewlog.setToolTip("Click to see an HTML rendering of your current log.")
        qa = title_tb.addAction(pixmaps.purr_logo.icon(), "About...", self._about_dialog.exec_)
        qa.setToolTip("<P>Click to see the About... dialog, which will tell you something about PURR.</P>")

        self.wdirframe = QFrame(cw)
        cwlo.addWidget(self.wdirframe)
        self.dirs_lo = QVBoxLayout(self.wdirframe)
        self.dirs_lo.setMargin(5)
        self.dirs_lo.setContentsMargins(5, 0, 5, 5)
        self.dirs_lo.setSpacing(0)
        self.wdirframe.setFrameStyle(QFrame.Box | QFrame.Raised)
        self.wdirframe.setLineWidth(1)

        ## Directories toolbar
        dirs_tb = QToolBar(self.wdirframe)
        dirs_tb.setToolButtonStyle(Qt.ToolButtonIconOnly)
        dirs_tb.setIconSize(QSize(16, 16))
        self.dirs_lo.addWidget(dirs_tb)
        label = QLabel("Monitoring directories:", dirs_tb)
        self._dirs_tip = """<P>PURR can monitor your working directories for new or updated files. If there's a checkmark
      next to the directory name in this list, PURR is monitoring it.</P>

      <P>If the checkmark is grey, PURR is monitoring things unobtrusively. When a new or updated file is detected in he monitored directory,
      it is quietly added to the list of files in the "New entry" window, even if this window is not currently visible.</P>

      <P>If the checkmark is black, PURR will be more obtrusive. Whenever a new or updated file is detected, the "New entry" window will
      pop up automatically. This is called "pouncing", and some people find it annoying.</P>
      """
        label.setToolTip(self._dirs_tip)
        label.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        dirs_tb.addWidget(label)

        # add directory list widget
        self.wdirlist = DirectoryListWidget(self.wdirframe)
        self.wdirlist.setToolTip(self._dirs_tip)
        QObject.connect(self.wdirlist, SIGNAL("directoryStateChanged"), self._changeWatchedDirState)
        self.dirs_lo.addWidget(self.wdirlist)
        # self.wdirlist.setMaximumSize(1000000,64)

        # add directory button
        add = dirs_tb.addAction(pixmaps.list_add.icon(), "Add", self._showAddDirectoryDialog)
        add.setToolTip("<P>Click to add another directory to be monitored.</P>")

        # remove directory button
        delbtn = dirs_tb.addAction(pixmaps.list_remove.icon(), "Remove", self.wdirlist.removeCurrent)
        delbtn.setEnabled(False)
        delbtn.setToolTip("<P>Click to removed the currently selected directory from the list.</P>")
        QObject.connect(self.wdirlist, SIGNAL("hasSelection"), delbtn.setEnabled)

        #    # qa = dirs_tb.addAction(pixmaps.blue_round_reload.icon(),"Rescan",self._forceRescan)
        #    # qa.setToolTip("Click to rescan the directories for any new or updated files.")
        #    self.wshownew = QCheckBox("show new files",dirs_tb)
        #    dirs_tb.addWidget(self.wshownew)
        #    self.wshownew.setCheckState(Qt.Checked)
        #    self.wshownew.setToolTip("""<P>If this is checked, the "New entry" window will pop up automatically whenever
        #  new or updated files are detected. If this is unchecked, the files will be added to the window quietly
        #        and unobtrusively; you can show the window manually by clicking on the "New entry..." button below.</P>""")
        #    self._dir_entries = {}

        cwlo.addSpacing(5)

        wlogframe = QFrame(cw)
        cwlo.addWidget(wlogframe)
        log_lo = QVBoxLayout(wlogframe)
        log_lo.setMargin(5)
        log_lo.setContentsMargins(5, 5, 5, 5)
        log_lo.setSpacing(0)
        wlogframe.setFrameStyle(QFrame.Box | QFrame.Raised)
        wlogframe.setLineWidth(1)

        # listview of log entries
        self.etw = LogEntryTree(cw)
        log_lo.addWidget(self.etw, 1)
        self.etw.header().setDefaultSectionSize(128)
        self.etw.header().setMovable(False)
        self.etw.setHeaderLabels(["date", "entry title", "comment"])
        if hasattr(QHeaderView, 'ResizeToContents'):
            self.etw.header().setResizeMode(0, QHeaderView.ResizeToContents)
        else:
            self.etw.header().setResizeMode(0, QHeaderView.Custom)
            self.etw.header().resizeSection(0, 120)
        self.etw.header().setResizeMode(1, QHeaderView.Interactive)
        self.etw.header().setResizeMode(2, QHeaderView.Stretch)
        self.etw.header().show()
        try:
            self.etw.setAllColumnsShowFocus(True)
        except AttributeError:
            pass;  # Qt 4.2+
        # self.etw.setShowToolTips(True)
        self.etw.setSortingEnabled(False)
        # self.etw.setColumnAlignment(2,Qt.AlignLeft|Qt.AlignTop)
        self.etw.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.etw.setRootIsDecorated(True)
        self.connect(self.etw, SIGNAL("itemSelectionChanged()"), self._entrySelectionChanged)
        self.connect(self.etw, SIGNAL("itemActivated(QTreeWidgetItem*,int)"), self._viewEntryItem)
        self.connect(self.etw, SIGNAL("itemContextMenuRequested"), self._showItemContextMenu)
        # create popup menu for data products
        self._archived_dp_menu = menu = QMenu(self)
        self._archived_dp_menu_title = QLabel()
        self._archived_dp_menu_title.setMargin(5)
        self._archived_dp_menu_title_wa = wa = QWidgetAction(self)
        wa.setDefaultWidget(self._archived_dp_menu_title)
        menu.addAction(wa)
        menu.addSeparator()
        menu.addAction(pixmaps.editcopy.icon(), "Restore file(s) from archived copy", self._restoreItemFromArchive)
        menu.addAction(pixmaps.editpaste.icon(), "Copy pathname of archived copy to clipboard",
                       self._copyItemToClipboard)
        self._current_item = None
        # create popup menu for entries
        self._entry_menu = menu = QMenu(self)
        self._entry_menu_title = QLabel()
        self._entry_menu_title.setMargin(5)
        self._entry_menu_title_wa = wa = QWidgetAction(self)
        wa.setDefaultWidget(self._entry_menu_title)
        menu.addAction(wa)
        menu.addSeparator()
        menu.addAction(pixmaps.filefind.icon(), "View this log entry", self._viewEntryItem)
        menu.addAction(pixmaps.editdelete.icon(), "Delete this log entry", self._deleteSelectedEntries)
        # buttons at bottom
        log_lo.addSpacing(5)
        btnlo = QHBoxLayout()
        log_lo.addLayout(btnlo)
        self.wnewbtn = QPushButton(pixmaps.filenew.icon(), "New entry...", cw)
        self.wnewbtn.setToolTip("Click to add a new log entry.")
        # self.wnewbtn.setFlat(True)
        self.wnewbtn.setEnabled(False)
        btnlo.addWidget(self.wnewbtn)
        btnlo.addSpacing(5)
        self.weditbtn = QPushButton(pixmaps.filefind.icon(), "View entry...", cw)
        self.weditbtn.setToolTip("Click to view or edit the selected log entry/")
        # self.weditbtn.setFlat(True)
        self.weditbtn.setEnabled(False)
        self.connect(self.weditbtn, SIGNAL("clicked()"), self._viewEntryItem)
        btnlo.addWidget(self.weditbtn)
        btnlo.addSpacing(5)
        self.wdelbtn = QPushButton(pixmaps.editdelete.icon(), "Delete", cw)
        self.wdelbtn.setToolTip("Click to delete the selected log entry or entries.")
        # self.wdelbtn.setFlat(True)
        self.wdelbtn.setEnabled(False)
        self.connect(self.wdelbtn, SIGNAL("clicked()"), self._deleteSelectedEntries)
        btnlo.addWidget(self.wdelbtn)
        # enable status line
        self.statusBar().show()
        Purr.progressMessage = self.message
        self._prev_msg = None
        # editor dialog for new entry
        self.new_entry_dialog = Purr.Editors.NewLogEntryDialog(self)
        self.connect(self.new_entry_dialog, SIGNAL("newLogEntry"), self._newLogEntry)
        self.connect(self.new_entry_dialog, SIGNAL("filesSelected"), self._addDPFiles)
        self.connect(self.wnewbtn, SIGNAL("clicked()"), self.new_entry_dialog.show)
        self.connect(self.new_entry_dialog, SIGNAL("shown"), self._checkPounceStatus)
        # entry viewer dialog
        self.view_entry_dialog = Purr.Editors.ExistingLogEntryDialog(self)
        self.connect(self.view_entry_dialog, SIGNAL("previous()"), self._viewPrevEntry)
        self.connect(self.view_entry_dialog, SIGNAL("next()"), self._viewNextEntry)
        self.connect(self.view_entry_dialog, SIGNAL("viewPath"), self._viewPath)
        self.connect(self.view_entry_dialog, SIGNAL("filesSelected"), self._addDPFilesToOldEntry)
        self.connect(self.view_entry_dialog, SIGNAL("entryChanged"), self._entryChanged)
        # saving a data product to an older entry will automatically drop it from the
        # new entry dialog
        self.connect(self.view_entry_dialog, SIGNAL("creatingDataProduct"),
                     self.new_entry_dialog.dropDataProducts)
        # resize selves
        width = Config.getint('main-window-width', 512)
        height = Config.getint('main-window-height', 512)
        self.resize(QSize(width, height))
        # create timer for pouncing
        self._timer = QTimer(self)
        self.connect(self._timer, SIGNAL("timeout()"), self._rescan)
        # create dict mapping index.html paths to entry numbers
        self._index_paths = {}

    def resizeEvent(self, ev):
        QMainWindow.resizeEvent(self, ev)
        sz = ev.size()
        Config.set('main-window-width', sz.width())
        Config.set('main-window-height', sz.height())

    def closeEvent(self, ev):
        if self._hide_on_close:
            ev.ignore()
            self.hide()
            self.new_entry_dialog.hide()
        else:
            if self.purrer:
                self.purrer.detach()
            return QMainWindow.closeEvent(self, ev)

    def message(self, msg, ms=2000, sub=False):
        if sub:
            if self._prev_msg:
                msg = ": ".join((self._prev_msg, msg))
        else:
            self._prev_msg = msg
        self.statusBar().showMessage(msg, ms)
        QCoreApplication.processEvents(QEventLoop.ExcludeUserInputEvents)

    def _changeWatchedDirState(self, pathname, watching):
        self.purrer.setWatchingState(pathname, watching)
        # update dialogs if dir list has changed
        if watching == Purr.REMOVED:
            self.purrpipes.pop(pathname)
            dirs = [path for path, state in self.purrer.watchedDirectories()]
            self.new_entry_dialog.setDefaultDirs(*dirs)
            self.view_entry_dialog.setDefaultDirs(*dirs)
        pass

    def _showAddDirectoryDialog(self):
        dd = str(QFileDialog.getExistingDirectory(self, "PURR: Add a directory to monitor")).strip()
        if dd:
            # adds a watched directory. Default initial setting of 'watching' is POUNCE if all
            # directories are in POUNCE state, or WATCHED otherwise.
            watching = max(Purr.WATCHED,
                           min([state for path, state in self.purrer.watchedDirectories()] or [Purr.WATCHED]))
            self.purrer.addWatchedDirectory(dd, watching)
            self.purrpipes[dd] = Purr.Pipe.open(dd)
            self.wdirlist.add(dd, watching)
            # update dialogs since dir list has changed
            dirs = [path for path, state in self.purrer.watchedDirectories()]
            self.new_entry_dialog.setDefaultDirs(*dirs)
            self.view_entry_dialog.setDefaultDirs(*dirs)

    def detachPurrlog(self):
        self.wdirlist.clear()
        self.purrer and self.purrer.detach()
        self.purrer = None

    def hasPurrlog(self):
        return bool(self.purrer)

    def attachPurrlog(self, purrlog, watchdirs=[]):
        """Attaches Purr to the given purrlog directory. Arguments are passed to Purrer object as is."""
        # check purrer stack for a Purrer already watching this directory
        dprint(1, "attaching to purrlog", purrlog)
        for i, purrer in enumerate(self.purrer_stack):
            if os.path.samefile(purrer.logdir, purrlog):
                dprint(1, "Purrer object found on stack (#%d),reusing\n", i)
                # found? move to front of stack
                self.purrer_stack.pop(i)
                self.purrer_stack.insert(0, purrer)
                # update purrer with watched directories, in case they have changed
                for dd in (watchdirs or []):
                    purrer.addWatchedDirectory(dd, watching=None)
                break
        # no purrer found, make a new one
        else:
            dprint(1, "creating new Purrer object")
            try:
                purrer = Purr.Purrer(purrlog, watchdirs)
            except Purr.Purrer.LockedError as err:
                # check that we could attach, display message if not
                QMessageBox.warning(self, "Catfight!", """<P><NOBR>It appears that another PURR process (%s)</NOBR>
          is already attached to <tt>%s</tt>, so we're not allowed to touch it. You should exit the other PURR
          process first.</P>""" % (err.args[0], os.path.abspath(purrlog)), QMessageBox.Ok, 0)
                return False
            except Purr.Purrer.LockFailError as err:
                QMessageBox.warning(self, "Failed to obtain lock", """<P><NOBR>PURR was unable to obtain a lock</NOBR>
          on directory <tt>%s</tt> (error was "%s"). The most likely cause is insufficient permissions.</P>""" % (
                os.path.abspath(purrlog), err.args[0]), QMessageBox.Ok, 0)
                return False
            self.purrer_stack.insert(0, purrer)
            # discard end of stack
            self.purrer_stack = self.purrer_stack[:3]
            # attach signals
            self.connect(purrer, SIGNAL("disappearedFile"),
                         self.new_entry_dialog.dropDataProducts)
            self.connect(purrer, SIGNAL("disappearedFile"),
                         self.view_entry_dialog.dropDataProducts)
        # have we changed the current purrer? Update our state then
        # reopen Purr pipes
        self.purrpipes = {}
        for dd, state in purrer.watchedDirectories():
            self.purrpipes[dd] = Purr.Pipe.open(dd)
        if purrer is not self.purrer:
            self.message("Attached to %s" % purrer.logdir, ms=10000)
            dprint(1, "current Purrer changed, updating state")
            # set window title
            path = Kittens.utils.collapseuser(os.path.join(purrer.logdir, ''))
            self.setWindowTitle("PURR - %s" % path)
            # other init
            self.purrer = purrer
            self.new_entry_dialog.hide()
            self.new_entry_dialog.reset()
            dirs = [path for path, state in purrer.watchedDirectories()]
            self.new_entry_dialog.setDefaultDirs(*dirs)
            self.view_entry_dialog.setDefaultDirs(*dirs)
            self.view_entry_dialog.hide()
            self.viewer_dialog.hide()
            self._viewing_ientry = None
            self._setEntries(self.purrer.getLogEntries())
            #      print self._index_paths
            self._viewer_timestamp = None
            self._updateViewer()
            self._updateNames()
            # update directory widgets
            self.wdirlist.clear()
            for pathname, state in purrer.watchedDirectories():
                self.wdirlist.add(pathname, state)
            # Reset _pounce to false -- this will cause checkPounceStatus() into a rescan
            self._pounce = False
            self._checkPounceStatus()
        return True

    def setLogTitle(self, title):
        if self.purrer:
            if title != self.purrer.logtitle:
                self.purrer.setLogTitle(title)
                self._updateViewer()
            self._updateNames()

    def _updateNames(self):
        self.wnewbtn.setEnabled(True)
        self.wviewlog.setEnabled(True)
        self._about_dialog.setText(self.about_message + """
      <P>Your current log resides in:<PRE>  <tt>%s</tt></PRE>To see your log in all its HTML-rendered
      glory, point your browser to <tt>index.html</tt> therein, or use the handy "View" button provided by PURR.</P>

      <P>Your current working directories are:</P>
      <P>%s</P>
      """ % (self.purrer.logdir,
             "".join(["<PRE>  <tt>%s</tt></PRE>" % name for name, state in self.purrer.watchedDirectories()])
             ))
        title = self.purrer.logtitle or "Unnamed log"
        self.title_editor.setText(title)
        self.viewer_dialog.setWindowTitle(title)

    def _showViewerDialog(self):
        self._updateViewer(True)
        self.viewer_dialog.show()

    @staticmethod
    def fileModTime(path):
        try:
            return os.path.getmtime(path)
        except:
            return None

    def _updateViewer(self, force=False):
        """Updates the viewer dialog.
        If dialog is not visible and force=False, does nothing.
        Otherwise, checks the mtime of the current purrer index.html file against self._viewer_timestamp.
        If it is newer, reloads it.
        """
        if not force and not self.viewer_dialog.isVisible():
            return
        # default text if nothing is found
        path = self.purrer.indexfile
        mtime = self.fileModTime(path)
        # return if file is older than our content
        if mtime and mtime <= (self._viewer_timestamp or 0):
            return
        busy = BusyIndicator()
        self.viewer_dialog.setDocument(path, empty=
        "<P>Nothing in the log yet. Try adding some log entries.</P>")
        self.viewer_dialog.reload()
        self.viewer_dialog.setLabel("""<P>Below is your full HTML-rendered log. Note that this is 
      only a bare-bones viewer, so only a limited set of links will work. 
      For a fully-functional view, use a proper HTML browser to look at the index file residing here:<BR>
      <tt>%s</tt></P>
      """ % self.purrer.indexfile)
        self._viewer_timestamp = mtime

    def _setEntries(self, entries):
        self.etw.clear()
        item = None
        self._index_paths = {}
        self._index_paths[os.path.abspath(self.purrer.indexfile)] = -1
        for i, entry in enumerate(entries):
            item = self._addEntryItem(entry, i, item)
            self._index_paths[os.path.abspath(entry.index_file)] = i
        self.etw.resizeColumnToContents(0)

    def _titleChanged(self):
        self.setLogTitle(str(self.title_editor.text()))

    def _checkPounceStatus(self):
        ## pounce = bool([ entry for entry in self._dir_entries.itervalues() if entry.watching ])
        pounce = bool([path for path, state in self.purrer.watchedDirectories() if state >= Purr.WATCHED])
        # rescan, if going from not-pounce to pounce
        if pounce and not self._pounce:
            self._rescan()
        self._pounce = pounce
        # start timer -- we need it running to check the purr pipe, anyway
        self._timer.start(2000)

    def _forceRescan(self):
        if not self.purrer:
            self.attachDirectory('.')
        self._rescan(force=True)

    def _rescan(self, force=False):
        if not self.purrer:
            return
        # if pounce is on, tell the Purrer to rescan directories
        if self._pounce or force:
            dps = self.purrer.rescan()
            if dps:
                filenames = [dp.filename for dp in dps]
                dprint(2, "new data products:", filenames)
                self.message("Pounced on " + ", ".join(filenames))
                if self.new_entry_dialog.addDataProducts(dps):
                    dprint(2, "showing dialog")
                    self.new_entry_dialog.show()
        # else read stuff from pipe
        for pipe in list(self.purrpipes.values()):
            do_show = False
            for command, show, content in pipe.read():
                if command == "title":
                    self.new_entry_dialog.suggestTitle(content)
                elif command == "comment":
                    self.new_entry_dialog.addComment(content)
                elif command == "pounce":
                    self.new_entry_dialog.addDataProducts(self.purrer.makeDataProducts(
                        [(content, not show)], unbanish=True))
                else:
                    print(("Unknown command received from Purr pipe: ", command))
                    continue
                do_show = do_show or show
            if do_show:
                self.new_entry_dialog.show()

    def _addDPFiles(self, *files):
        """callback to add DPs corresponding to files."""
        # quiet flag is always true
        self.new_entry_dialog.addDataProducts(self.purrer.makeDataProducts(
            [(file, True) for file in files], unbanish=True, unignore=True))

    def _addDPFilesToOldEntry(self, *files):
        """callback to add DPs corresponding to files."""
        # quiet flag is always true
        self.view_entry_dialog.addDataProducts(self.purrer.makeDataProducts(
            [(file, True) for file in files], unbanish=True, unignore=True))

    def _entrySelectionChanged(self):
        selected = [item for item in self.etw.iterator(self.etw.Iterator.Selected) if item._ientry is not None]
        self.weditbtn.setEnabled(len(selected) == 1)
        self.wdelbtn.setEnabled(bool(selected))

    def _viewEntryItem(self, item=None, *dum):
        """Pops up the viewer dialog for the entry associated with the given item.
        If 'item' is None, looks for a selected item in the listview.
        The dum arguments are for connecting this to QTreeWidget signals such as doubleClicked().
        """
        # if item not set, look for selected items in listview. Only 1 must be selected.
        select = True
        if item is None:
            selected = [item for item in self.etw.iterator(self.etw.Iterator.Selected) if item._ientry is not None]
            if len(selected) != 1:
                return
            item = selected[0]
            select = False;  # already selected
        else:
            # make sure item is open -- the click will cause it to close
            self.etw.expandItem(item)
        # show dialog
        ientry = getattr(item, '_ientry', None)
        if ientry is not None:
            self._viewEntryNumber(ientry, select=select)

    def _viewEntryNumber(self, ientry, select=True):
        """views entry #ientry. Also selects entry in listview if select=True"""
        # pass entry to viewer dialog
        self._viewing_ientry = ientry
        entry = self.purrer.entries[ientry]
        busy = BusyIndicator()
        self.view_entry_dialog.viewEntry(entry,
                                         prev=ientry > 0 and self.purrer.entries[ientry - 1],
                                         next=ientry < len(self.purrer.entries) - 1 and self.purrer.entries[ientry + 1])
        self.view_entry_dialog.show()
        # select entry in listview
        if select:
            self.etw.clearSelection()
            self.etw.setItemSelected(self.etw.topLevelItem(ientry), True)

    def _viewPrevEntry(self):
        if self._viewing_ientry is not None and self._viewing_ientry > 0:
            self._viewEntryNumber(self._viewing_ientry - 1)

    def _viewNextEntry(self):
        if self._viewing_ientry is not None and self._viewing_ientry < len(self.purrer.entries) - 1:
            self._viewEntryNumber(self._viewing_ientry + 1)

    def _viewPath(self, path):
        num = self._index_paths.get(os.path.abspath(path), None)
        if num is None:
            return
        elif num == -1:
            self.view_entry_dialog.hide()
            self._showViewerDialog()
        else:
            self._viewEntryNumber(num)

    def _showItemContextMenu(self, item, point, col):
        """Callback for contextMenuRequested() signal. Pops up item menu, if defined"""
        menu = getattr(item, '_menu', None)
        if menu:
            settitle = getattr(item, '_set_menu_title', None)
            if settitle:
                settitle()
            # self._current_item tells callbacks what item the menu was referring to
            point = self.etw.mapToGlobal(point)
            self._current_item = item
            self.etw.clearSelection()
            self.etw.setItemSelected(item, True)
            menu.exec_(point)
        else:
            self._current_item = None

    def _copyItemToClipboard(self):
        """Callback for item menu."""
        if self._current_item is None:
            return
        dp = getattr(self._current_item, '_dp', None)
        if dp and dp.archived:
            path = dp.fullpath.replace(" ", "\\ ")
            QApplication.clipboard().setText(path, QClipboard.Clipboard)
            QApplication.clipboard().setText(path, QClipboard.Selection)

    def _restoreItemFromArchive(self):
        """Callback for item menu."""
        if self._current_item is None:
            return
        dp = getattr(self._current_item, '_dp', None)
        if dp and dp.archived:
            dp.restore_from_archive(parent=self)

    def _deleteSelectedEntries(self):
        remaining_entries = []
        del_entries = list(self.etw.iterator(self.etw.Iterator.Selected))
        remaining_entries = list(self.etw.iterator(self.etw.Iterator.Unselected))
        if not del_entries:
            return
        hide_viewer = bool([item for item in del_entries if self._viewing_ientry == item._ientry])
        del_entries = [self.purrer.entries[self.etw.indexOfTopLevelItem(item)] for item in del_entries]
        remaining_entries = [self.purrer.entries[self.etw.indexOfTopLevelItem(item)] for item in remaining_entries]
        # ask for confirmation
        if len(del_entries) == 1:
            msg = """<P><NOBR>Permanently delete the log entry</NOBR> "%s"?</P>""" % del_entries[0].title
            if del_entries[0].dps:
                msg += """<P>%d data product(s) saved with this
                  entry will be deleted as well.</P>""" % len(del_entries[0].dps)
        else:
            msg = """<P>Permanently delete the %d selected log entries?</P>""" % len(del_entries)
            ndp = 0
            for entry in del_entries:
                ndp += len([dp for dp in entry.dps if not dp.ignored])
            if ndp:
                msg += """<P>%d data product(s) saved with these entries will be deleted as well.</P>""" % ndp
        if QMessageBox.warning(self, "Deleting log entries", msg,
                               QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
            return
        if hide_viewer:
            self.view_entry_dialog.hide()
        # reset entries in purrer and in our log window
        self._setEntries(remaining_entries)
        self.purrer.deleteLogEntries(del_entries)
        #    self.purrer.setLogEntries(remaining_entries)
        # log will have changed, so update the viewer
        self._updateViewer()
        # delete entry files
        for entry in del_entries:
            entry.remove_directory()

    def _addEntryItem(self, entry, number, after):
        item = entry.tw_item = QTreeWidgetItem(self.etw, after)
        timelabel = self._make_time_label(entry.timestamp)
        item.setText(0, timelabel)
        item.setText(1, " " + (entry.title or ""))
        item.setToolTip(1, entry.title)
        if entry.comment:
            item.setText(2, " " + entry.comment.split('\n')[0])
            item.setToolTip(2, "<P>" + entry.comment.replace("<", "&lt;").replace(">", "&gt;"). \
                            replace("\n\n", "</P><P>").replace("\n", "</P><P>") + "</P>")
        item._ientry = number
        item._dp = None
        item._menu = self._entry_menu
        item._set_menu_title = lambda: self._entry_menu_title.setText('"%s"' % entry.title)
        # now make subitems for DPs
        subitem = None
        for dp in entry.dps:
            if not dp.ignored:
                subitem = self._addDPSubItem(dp, item, subitem)
        self.etw.collapseItem(item)
        self.etw.header().headerDataChanged(Qt.Horizontal, 0, 2)
        return item

    def _addDPSubItem(self, dp, parent, after):
        item = QTreeWidgetItem(parent, after)
        item.setText(1, dp.filename)
        item.setToolTip(1, dp.filename)
        item.setText(2, dp.comment or "")
        item.setToolTip(2, dp.comment or "")
        item._ientry = None
        item._dp = dp
        item._menu = self._archived_dp_menu
        item._set_menu_title = lambda: self._archived_dp_menu_title.setText(os.path.basename(dp.filename))
        return item

    def _make_time_label(self, timestamp):
        return time.strftime("%b %d %H:%M", time.localtime(timestamp))

    def _newLogEntry(self, entry):
        """This is called when a new log entry is created"""
        # add entry to purrer
        self.purrer.addLogEntry(entry)
        # add entry to listview if it is not an ignored entry
        # (ignored entries only carry information about DPs to be ignored)
        if not entry.ignore:
            if self.etw.topLevelItemCount():
                lastitem = self.etw.topLevelItem(self.etw.topLevelItemCount() - 1)
            else:
                lastitem = None
            self._addEntryItem(entry, len(self.purrer.entries) - 1, lastitem)
            self._index_paths[os.path.abspath(entry.index_file)] = len(self.purrer.entries) - 1
        # log will have changed, so update the viewer
        if not entry.ignore:
            self._updateViewer()
            self.show()

    def _entryChanged(self, entry):
        """This is called when a log entry is changed"""
        # resave the log
        self.purrer.save()
        # redo entry item
        if entry.tw_item:
            number = entry.tw_item._ientry
            entry.tw_item = None
            self.etw.takeTopLevelItem(number)
            if number:
                after = self.etw.topLevelItem(number - 1)
            else:
                after = None
            self._addEntryItem(entry, number, after)
        # log will have changed, so update the viewer
        self._updateViewer()

    def _regenerateLog(self):
        if QMessageBox.question(self.viewer_dialog, "Regenerate log", """<P><NOBR>Do you really want to regenerate the
      entire</NOBR> log? This can be a time-consuming operation.</P>""",
                                QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
            return
        self.purrer.save(refresh=True)
        self._updateViewer()



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
#
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import getopt
import os
import sys

__doc__ = """This module is defines variables for changing the looks
of charts. All the variables can be changed either via environment
variable PYCHART_OPTIONS or via the command-line options.

The value of PYCHART_OPTIONS should be a sequence of var=val separated
by space.  Below is an example, which tells Pychart to write to file
foo.pdf and use Times-Roman as the default font.

PYCHART_OPTIONS="output=foo.pdf font-family=Times"

The summary of attributes that can be set via PYCHART_OPTIONS follows:

output=FILENAME (default: stdout)

    Set the output file name.

format=[ps|pdf|pdf-uncompressed|png|x11|svg] (default: ps)

    Set the output file format.

font-family=NAME (default: Helvetica)

    Set the default font to be used by texts.

font-size=N (default: 9)

    Set the default font to be used by texts.
line-width=X (default: 0.4)

    Set the default line width, in points.  See also
    pychart.line_style.

scale=X (default: 1.0)

    Set the scaling factor.  The default is 1.0.

color=[yes|no] (default: no)

    If yes, Pychart colorizes default object attributes.

You can also set these variables by calling theme.get_options.
"""

use_color = 0
scale_factor = 1
output_format = None  # "ps", "pdf", "png", "x11", or "svg"
compress_output = 1
output_file = ""

default_font_family = "Helvetica"
default_font_size = 9
default_line_height = None
default_font_halign = "L"
default_font_valign = "B"
default_font_angle = 0
default_line_width = 0.4

debug_level = 1
delta_bounding_box = [-3, -3, 3, 3]
bounding_box = {}

creator = None
creation_date = None
author = None
title = None
aux_comments = None


def parse_yesno(str):
    if str in ("yes", "true", "1"):
        return 1
    else:
        return 0


def parse_bounding_box(arg):
    global delta_bounding_box, bounding_box

    l = arg.split(",")
    if len(l) != 4:
        raise ValueError("Need to specify margin=LEFT,BOTTOM,RIGHT,TOP")
    for i in range(0, 4):
        val = l[i].strip()
        if val[0] == '+':
            delta_bounding_box[i] = int(val[1:])
        elif val[0] == '-':
            delta_bounding_box[i] = int(val[1:])
        else:
            bounding_box[i] = int(val)


def adjust_bounding_box(bbox):
    """Adjust the bounding box as specified by user.
    Returns the adjusted bounding box.

    - bbox: Bounding box computed from the canvas drawings.
    It must be a four-tuple of numbers.
    """
    for i in range(0, 4):
        if i in bounding_box:
            bbox[i] = bounding_box[i]
        else:
            bbox[i] += delta_bounding_box[i]
    return bbox


def parse_option(opt, arg):
    global use_color, scale_factor, margin
    global output_format, output_file, compress_output
    global default_font_family, default_font_size
    global default_line_height
    global default_line_width, debug_level
    global creator, creation_date, author, title, aux_comments
    if opt == "format":
        if arg in ("ps", "eps"):
            output_format = "ps"
        elif arg == "png":
            output_format = "png"
        elif arg == "svg":
            output_format = "svg"
        elif arg == "x11":
            output_format = "x11"
        elif arg == "pdf-uncompressed":
            output_format = "pdf"
            compress_output = 0
        elif arg in ("pdf-compressed", "pdf"):
            output_format = "pdf"
            compress_output = 1
        else:
            raise ValueError("Unknown output option: " + str(arg))
    elif opt == "output":
        output_file = arg
    elif opt == "color":
        use_color = 1
    elif opt == "scale":
        scale_factor = float(arg)
    elif opt == "bbox":
        parse_bounding_box(arg)
    elif opt == "font-family":
        default_font_family = arg
    elif opt == "font-size":
        default_font_size = float(arg)
        default_line_height = float(arg)
    elif opt == "line-width":
        default_line_width = float(arg)
    elif opt == "debug-level":
        debug_level = int(arg)
    elif opt == "author":
        author = arg
    elif opt == "creator":
        creator = arg
    elif opt == "creation_date":
        creation_date = arg
    elif opt == "title":
        title = arg
    elif opt == "aux_comments":
        aux_comments = arg
    else:
        raise getopt.GetoptError("Unknown option: " + opt + " " + arg)


if "PYCHART_OPTIONS" in os.environ:
    for opt in os.environ["PYCHART_OPTIONS"].split():
        opt, arg = opt.split("=")
        parse_option(opt, arg)

hooks = []


def add_reinitialization_hook(proc):
    global hooks
    hooks.append(proc)
    proc()


def usage():
    print(("Usage: %s [options..]" % sys.argv[0]))
    print("""
    --scale=X: Set the scaling factor to X (default: 1.0).
    --format=[ps|png|pdf|x11|svg]: Set the output format (default: ps).
    --font-family=NAME: Set the default font family (default: Helvetica).
    --font-size=NAME: Set the default font size (default: 9pts).
    --line-width=NAME: Set the default line width (default: 0.4).
    --debug-level=N: Set the messaging verbosity (default: 0).
    --bbox=LEFT,BOTTOM,RIGHT,TOP: Specifies the amount of space (in PS points) to be left in the edges of the picture (default: -1,-1,+1,+1).
    """)


def reinitialize():
    """This procedure must be called after setting variables in
    the |theme| module. This procedure propagates the new values of
    the theme variables to other modules that depend on their values."""
    for proc in hooks:
        proc()


def get_options(argv=None):
    """This procedure takes a list of command line arguments in <argv>
    and parses
    options. It returns the non-parsed portion of <argv>. Parameter
    <argv> can be
    omitted, in which case its value defaults to |sys.argv[1:]|.
    The options supported are: "|--format=[ps,png,pdf,x11,svg]|",
    "|--output=|<file>", "|--color=[yes,no]|"
    "|--scale=|<X>", "|--font-family=|<name>", "|--font-size=|<X>",
    "|--line-width=|<X>",
    "|--debug-level=|<N>", "|bbox=|<left,bottom,right,top>".
    The below code shows an example.

#!/usr/bin/python
from pychart import *
args = theme.get_options()
ar = area.T(...)
...
    """
    if not argv:
        argv = sys.argv[1:]
    try:
        opts, args = getopt.getopt(argv, "d:co:f:",
                                   ["format=", "output=", "color=",
                                    "scale=", "font-family=", "font-size=",
                                    "line-width=", "debug-level=",
                                    "title=", "author=", "creation_date=",
                                    "creator=",
                                    "bbox="])
    except getopt.GetoptError as e:
        print(e)
        usage()
        raise
    else:
        for opt, arg in opts:
            if opt == "-d":
                parse_option("debug-level", arg)
            elif opt == "-c":
                parse_option("color", None)
            elif opt == "-o":
                parse_option("output", arg)
            elif opt == "-f":
                parse_option("format", arg)
            else:
                parse_option(opt[2:], arg)
        reinitialize()
        return args



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from types import *

from . import area_doc
from . import axis
from . import canvas
from . import category_coord
from . import chart_object
from . import coord
from . import fill_style
from . import legend
from . import line_style
from . import linear_coord
from . import theme
from .pychart_types import *

_dummy_legend = legend.T()


def range_doc(t):
    u = t.upper()

    return """Specifies the range of %s values that are displayed in the
    chart. IF the value is None, both the values are computed
    automatically from the samples.  Otherwise, the value must be a
    tuple of format (MIN, MAX). MIN and MAX must be either None or a
    number. If None, the value is computed automatically from the
    samples. For example, if %s_range = (None,5), then the minimum %s
    value is computed automatically, but the maximum %s value is fixed
    at 5.""" % (u, t, u, u)


_keys = {
    "loc": (CoordType, (0, 0),
            """The location of the bottom-left corner of the chart.
            None means to use the default value.
@cindex location (- of chart)
"""),
    "size": (CoordType, (120, 110),
             """The size of the chart-drawing area, excluding axis labels,
             legends, tick marks, etc. None means to use the default value.
@cindex size!- of area
             """),
    "bg_style": (fill_style.T, None, "Background fill-pattern."),
    "border_line_style": (line_style.T, None, "Line style of the outer frame of the chart."),
    "x_coord":
        (coord.T, linear_coord.T(),
         """Set the X coordinate system.""",
         """A linear coordinate system."""),
    "y_coord": (coord.T, linear_coord.T(),
                "Set the Y coordinate system.",
                """A linear coordinate system."""),
    "x_range": (CoordType, None, range_doc("x")),
    "y_range": (CoordType, None, range_doc("y")),
    "x_axis": (axis.X, None, "The X axis. <<axis>>."),
    "x_axis2": (axis.X, None, """The second X axis. This axis should be non-None either when you want to display plots with two distinct domains or when
    you just want to display two axes at the top and bottom of the chart.
    <<axis>>"""),
    "y_axis": (axis.Y, None, "The Y axis. <<axis>>."),
    "y_axis2": (axis.Y, None,
                """The second Y axis. This axis should be non-None either when you want to display plots with two distinct ranges or when
                you just want to display two axes at the left and right of the chart. <<axis>>"""),
    "x_grid_style": (line_style.T, None,
                     """The style of horizontal grid lines.
@cindex grid lines"""),
    "y_grid_style": (line_style.T, line_style.gray70_dash3,
                     "The style of vertical grid lines."),
    "x_grid_interval": (IntervalType, None,
                        """The horizontal grid-line interval.
                        A numeric value
                        specifies the interval at which
                        lines are drawn. If value is a function, it
                        takes two arguments, (MIN, MAX), that tells
                        the minimum and maximum values found in the
                        sample data. The function should return a list
                        of values at which lines are drawn."""),
    "y_grid_interval": (IntervalType, None,
                        "The vertical grid-line interval. See also x_grid_interval"),
    "x_grid_over_plot": (bool, False,
                         "If True, grid lines are drawn over plots. Otherwise, plots are drawn over grid lines."),
    "y_grid_over_plot": (IntType, False, "See x_grid_over_plot."),
    "legend": (legend.T, _dummy_legend, "The legend of the chart.",
               """a legend is by default displayed
               in the right-center of the chart. If you don't want to draw
               a legend, you should explicitly set None to this attribute."""),
}


class T(chart_object.T):
    keys = _keys
    __doc__ = area_doc.doc

    def __init__(self, **args):
        chart_object.T.init(self, args)
        self.__plots = []

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED
    def x_pos(self, xval):
        "Return the x position (on the canvas) corresponding to XVAL."
        off = self.x_coord.get_canvas_pos(self.size[0], xval,
                                          self.x_range[0], self.x_range[1])
        return self.loc[0] + off

    def y_pos(self, yval):
        "Return the y position (on the canvas) corresponding to YVAL."
        off = self.y_coord.get_canvas_pos(self.size[1], yval,
                                          self.y_range[0], self.y_range[1])
        return self.loc[1] + off

    def x_tic_points(self, interval):
        "Return the list of X values for which tick marks and grid lines are drawn."
        if type(interval) == FunctionType:
            return interval(*self.x_range)

        return self.x_coord.get_tics(self.x_range[0], self.x_range[1], interval)

    def y_tic_points(self, interval):
        "Return the list of Y values for which tick marks and grid lines are drawn."
        if type(interval) == FunctionType:
            return interval(*self.y_range)

        return self.y_coord.get_tics(self.y_range[0], self.y_range[1], interval)

    def __draw_x_grid_and_axis(self, can):
        if self.x_grid_style:
            for i in self.x_tic_points(self.x_grid_interval):
                x = self.x_pos(i)
                if x > self.loc[0]:
                    can.line(self.x_grid_style,
                             x, self.loc[1], x, self.loc[1] + self.size[1])
        if self.x_axis:
            self.x_axis.draw(self, can)
        if self.x_axis2:
            self.x_axis2.draw(self, can)

    def __draw_y_grid_and_axis(self, can):
        if self.y_grid_style:
            for i in self.y_tic_points(self.y_grid_interval):
                y = self.y_pos(i)
                if y > self.loc[1]:
                    can.line(self.y_grid_style,
                             self.loc[0], y,
                             self.loc[0] + self.size[0], y)
        if self.y_axis:
            self.y_axis.draw(self, can)
        if self.y_axis2:
            self.y_axis2.draw(self, can)

    def __get_data_range(self, r, which, coord, interval):
        if isinstance(coord, category_coord.T):
            # This info is unused for the category coord type.
            # So I just return a random value.
            return ((0, 0), 1)

        r = r or (None, None)

        if len(self.__plots) == 0:
            raise ValueError("No chart to draw, and no data range specified.\n")
        dmin, dmax = 999999, -999999

        for plot in self.__plots:
            this_min, this_max = plot.get_data_range(which)
            dmin = min(this_min, dmin)
            dmax = max(this_max, dmax)

        if interval and type(interval) == FunctionType:
            tics = interval(*(dmin, dmax))
            dmin = tics[0]
            dmax = tics[len(tics) - 1]
        else:
            dmin, dmax, interval = coord.get_min_max(dmin, dmax, interval)

        if r[0] != None:
            dmin = r[0]
        if r[1] != None:
            dmax = r[1]
        return ((dmin, dmax), interval)

    def draw(self, can=None):
        "Draw the charts."

        if can == None:
            can = canvas.default_canvas()

        assert self.check_integrity()

        for plot in self.__plots:
            plot.check_integrity()

        self.x_range, self.x_grid_interval = \
            self.__get_data_range(self.x_range, 'X',
                                  self.x_coord,
                                  self.x_grid_interval)

        self.y_range, self.y_grid_interval = \
            self.__get_data_range(self.y_range, 'Y',
                                  self.y_coord,
                                  self.y_grid_interval)

        can.rectangle(self.border_line_style, self.bg_style,
                      self.loc[0], self.loc[1],
                      self.loc[0] + self.size[0], self.loc[1] + self.size[1])

        if not self.x_grid_over_plot:
            self.__draw_x_grid_and_axis(can)

        if not self.y_grid_over_plot:
            self.__draw_y_grid_and_axis(can)

        clipbox = theme.adjust_bounding_box([self.loc[0], self.loc[1],
                                             self.loc[0] + self.size[0],
                                             self.loc[1] + self.size[1]])

        can.clip(clipbox[0], clipbox[1],
                 clipbox[2], clipbox[3])

        for plot in self.__plots:
            plot.draw(self, can)

        can.endclip()

        if self.x_grid_over_plot:
            self.__draw_x_grid_and_axis(can)
        if self.y_grid_over_plot:
            self.__draw_y_grid_and_axis(can)

        if self.legend == _dummy_legend:
            self.legend = legend.T()

        if self.legend:
            legends = []
            for plot in self.__plots:
                entry = plot.get_legend_entry()
                if entry == None:
                    pass
                elif type(entry) != ListType:
                    legends.append(entry)
                else:
                    for e in entry:
                        legends.append(e)
            self.legend.draw(self, legends, can)

    def plots(self):
        return self.__plots

    def add_plot(self, *plots):
        "Add PLOTS... to the area."
        self.__plots.extend(plots)



# -*- coding: utf-8 -*-
Version = "1.1"

# some global constants

# these are the different watching modes for the directories
REMOVED = -1;  # directory removed from watchlist
UNWATCHED = 0;  # directory in list, but not watched
WATCHED = 1;  # directory watched for new files quietly
POUNCE = 2;  # directory watched loudly

import os.path

# init debug printing
import Kittens.utils

verbosity = Kittens.utils.verbosity(name="purr")
dprint = verbosity.dprint
dprintf = verbosity.dprintf

import Kittens.pixmaps

pixmaps = Kittens.pixmaps.PixmapCache("Purr")

import Kittens.config

ConfigFile = Kittens.config.DualConfigParser("purr.conf")
Config = Kittens.config.SectionParser(ConfigFile, "Purr")

from .LogEntry import DataProduct, LogEntry
from .Purrer import Purrer
from . import Parsers


# if GUI is enabled, this will be overwritten by an object that
# sets a busy cursor on instantiation, and restores the cursor when killed
class BusyIndicator(object):
    pass


def canonizePath(path):
    """Returns the absolute, normalized, real path  of something.
    This takes care of symbolic links, redundant separators, etc."""
    return os.path.abspath(os.path.normpath(os.path.realpath(path)))


def progressMessage(msg, sub=False):
    """shows a progress message. If a GUI is available, this will be redefined by the GUI.
    If sub is true, message is a sub-message of previous one."""
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# loading file into karma:
#   ksend -add_connection unix 9975 multi_array kvis -load ARRAY:FILE0 img4096.fits
# ls /tmp/.KARMA-connections for port number
from PyQt4.Qt import QApplication
import sys


def trace_lines(frame, event, arg):
    if event == "line":
        print(("%s:%d" % (frame.f_code.co_filename, frame.f_lineno)))
    return trace_lines


# runs Purr standalone
if __name__ == "__main__":
    print("Welcome to PURR!")

    # parse options is the first thing we should do
    from optparse import OptionParser

    usage = "usage: %prog [options] <directories to watch>"
    parser = OptionParser(usage=usage)
    parser.add_option("-d", "--debug", dest="verbose", type="string", action="append", metavar="Context=Level",
                      help="(for debugging Python code) sets verbosity level of the named Python context. May be used multiple times.")
    parser.add_option("--trace", dest="trace", action="store_true",
                      help="(for debugging Python code) enables line tracing of Python statements")
    (options, rem_args) = parser.parse_args()

    if options.trace:
        sys.settrace(trace_lines)

    print("Please wait a second while the GUI starts up.")

    import sys
    import signal
    import os
    import os.path

    from PyQt4.Qt import *

    import Purr
    import Purr.MainWindow
    import Purr.Render
    import Purr.Startup

    app = QApplication(sys.argv)
    app.setDesktopSettingsAware(True)

    #  splash = QSplashScreen(Purr.pixmaps.purr_logo.pm())
    #  splash.showMessage("PURR!")
    #  splash.show()

    purrwin = Purr.MainWindow.MainWindow(None)

    try:
        if not Purr.Startup.startWizard(rem_args, purrwin):
            print("Cancelled by user")
            os._exit(1)
    except Purr.Startup.Error as err:
        print((err.error_message))
        os._exit(1)


    # handle SIGINT
    def sigint_handler(sig, stackframe):
        print("Caught Ctrl+C, PURR exiting...")
        purrwin.detachPurrlog()
        app.quit()


    signal.signal(signal.SIGINT, sigint_handler)

    app.exec_()

    os._exit(0)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import bar_plot_doc
from . import chart_object
from . import fill_style
from . import legend
from . import line_style
from . import pychart_util
from . import theme
from .pychart_types import *

fill_styles = None

_keys = {
    "direction": (StringType, "vertical",
                  """The direction the growth of the bars. The value is either 'horizontal'
                  or 'vertical'."""),
    "data": (AnyType, None, """Specifes data points. Unlike other types
    of charts, the "hcol"th column of the data must be a sequence of
    numbers, not just a single number. See also the description of
    "hcol"."""
             ),
    "data_label_offset": (CoordType, (0, 5),
                          "The location of data labels relative to the sample point. See also attribute data_label_format."),

    "data_label_format": (FormatType, None, """The
                          format string for the label displayed besides each
                          bar.  It can be a `printf' style format
                          string, or a two-parameter function that
                          takes (x,y) values and returns a string. """
                          + pychart_util.string_desc),

    "label": (StringType, "???", pychart_util.label_desc),
    "bcol": (IntType, 0,
             """Specifies the column from which base values (i.e., X values when attribute "direction" is "vertical", Y values otherwise) are extracted.
The
             combination of "data", "bcol", and "hcol" attributes defines
             the set of boxes drawn by this chart.
             See also the descriptions of the 'bcol' and 'data' attributes.
             """),
    "hcol": (IntType, 1,
             """The column from which the base and height of
             bars are extracted. See the below example:
              
@example
              d = [[5,[10,15,22]], [7,[22,23,5,10]], [8,[25,3]]]
              p = interval_bar_plot.T(data = d, bcol = 0, hcol = 1)
@end example

              Here, three sequence of bars will be drawn.
              The X locations of the bars
              will be 5, 7, and 8. For example, at location X=7,
              three bars are drawn,
              one corresponding to Y values of 22 to 45 (=22+23),
              and the second one for values 45 to 50, and the third one
              for values 50 to 60. The line and fill styles of the bars
              are picked in a round-robin fashion
              from attributes "line_styles" and
              "fill_styles".
             """),
    "line_styles": (ListType, [line_style.default, None],
                    """The list of line styles for bars.
                    The style of each bar is chosen in a round-robin fashion, if the
                    number of elements in "line_styles" is smaller than
                    actual number of boxes."""),
    "fill_styles": (ListType, [lambda: next(fill_styles), None],   # type: ignore
                    """List of fill styles for bars.
                    The style of each bar is chosen in a round-robin fashion, if the
                    number of elements in "line_styles" is smaller than
                    actual number of boxes.
                    If this attribute is omitted,
                    a style is picked from standard styles round-robin. <<fill_style>>."""),
    "cluster": (TupleType, (0, 1), """This attribute is used to
    cluster multiple bar plots side by side in a single chart.
    The value should be a tuple of two integers. The second value should be equal to the total number of bar plots in the chart. The first value should be the relative position of this chart; 0 places this chart the leftmost, and N-1
    (where N is the 2nd value of this attribute) places this chart the rightmost. Consider the below example:

@example
    a = area.T(...)
    p1 = interval_bar_plot.T(data = [[1, [20,10]][2,[30,5]]], cluster=(0,2))
    p2 = interval_bar_plot.T(data = [[1,[25,11,2]],[2,[10,5,3]]], cluster=(1,2))
    a.add_plot(p1, p2)
    a.draw()
@end example

    In this example, one group of bars will be drawn side-by-side at
    position x=1.
    Other two bars will be drawn side by side at position x=2.
    See also the description of attribute "cluster" for bar_plot.T.
    """),
    "width": (UnitType, 5, """Width of each box. The unit is in points.
@cindex width!- of interval bar plot
@cindex size!- interval bar plot
"""),
    "cluster_sep": (UnitType, 0, """The separation between
    clustered boxes. The unit is points."""),
    "stack_on": (AnyType, None,
                 "The value must be either None or bar_plot.T. If not None, bars of this plot are stacked on top of another bar plot."),
}


class T(chart_object.T):
    __doc__ = bar_plot_doc.doc
    keys = _keys

    def check_integrity(self):
        chart_object.T.check_integrity(self)

    def get_value(self, bval):
        for pair in self.data:
            if pair[self.bcol] == bval:
                return pair[self.hcol]
        raise ValueError(str(bval) + ": can't find the xval")

    def __get_data_range(self, col):
        gmin = 99999999
        gmax = -99999999
        for item in self.data:
            seq = item[col]
            if seq[0] < gmin: gmin = seq[0]
            max = 0
            for v in seq:
                max += v
            if max > gmax: gmax = max
        return (gmin, gmax)

    def get_data_range(self, which):
        if self.direction == 'vertical':
            if which == 'X':
                return pychart_util.get_data_range(self.data, self.bcol)
            else:
                return self.__get_data_range(self.hcol)
        else:
            assert self.direction == 'horizontal'
            if which == 'Y':
                return pychart_util.get_data_range(self.data, self.bcol)
            else:
                return self.__get_data_range(self.hcol)

    def get_style(self, nth):
        line_style = self.line_styles[nth % len(self.line_styles)]
        fill_style = self.fill_styles[nth % len(self.fill_styles)]
        return (line_style, fill_style)

    def draw_vertical(self, ar, can):
        for pair in self.data:
            xval = pair[self.bcol]
            yvals = pychart_util.get_sample_val(pair, self.hcol)

            if None in (xval, yvals): continue

            ybot = 0

            totalWidth = (self.width + self.cluster_sep) * self.cluster[1] - self.cluster_sep
            first_x = ar.x_pos(xval) - totalWidth / 2.0
            this_x = first_x + (self.width + self.cluster_sep) * self.cluster[0] - self.cluster_sep

            cury = yvals[0]
            n = 0

            for yval in yvals[1:]:
                (line_style, fill_style) = self.get_style(n)
                can.rectangle(line_style, fill_style,
                              this_x, ar.y_pos(cury), this_x + self.width,
                              ar.y_pos(cury + yval))
                cury += yval
                n += 1

                if self.data_label_format:
                    can.show(this_x + self.width / 2.0 + self.data_label_offset[0],
                             ar.y_pos(cury) + self.data_label_offset[1],
                             "/hC" + pychart_util.apply_format(self.data_label_format,
                                                               (pair[self.bcol], pair[self.hcol]), 1))

    def draw_horizontal(self, ar, can):
        for pair in self.data:
            yval = pair[self.bcol]
            xvals = pychart_util.get_sample_val(pair, self.hcol)

            if None in (xvals, yval): continue

            totalWidth = (self.width + self.cluster_sep) * self.cluster[1] - self.cluster_sep
            first_y = ar.y_pos(yval) - totalWidth / 2.0
            this_y = first_y + (self.width + self.cluster_sep) * self.cluster[0] - self.cluster_sep

            curx = xvals[0]
            n = 0
            for xval in xvals[1:]:
                line_style, fill_style = self.get_style(n)
                can.rectangle(line_style, fill_style,
                              ar.x_pos(curx), this_y,
                              ar.x_pos(curx + xval), this_y + self.width)
                curx += xval
                n += 1

    def get_legend_entry(self):
        if self.label:
            return legend.Entry(line_style=self.line_styles[0],
                                fill_style=self.fill_styles[0],
                                label=self.label)
        return None

    def draw(self, ar, can):
        assert chart_object.T.check_integrity(self)
        can.clip(ar.loc[0], ar.loc[1],
                 ar.loc[0] + ar.size[0], ar.loc[1] + ar.size[1])

        if self.direction == "vertical":
            self.draw_vertical(ar, can)
        else:
            self.draw_horizontal(ar, can)

        can.endclip()


def init():
    global fill_styles
    fill_styles = fill_style.standards.iterate()


theme.add_reinitialization_hook(init)



# -*- coding: utf-8 -*-

import traceback

from Purr.Render import DefaultRenderer, dprintf


class CachingRenderer(DefaultRenderer):
    """This is a base class for "expensive" renderers.
    It maintains a cache of rendered HTML code, and only invokes the renderer if the cache is out of date."""

    def _cacheFileExtension(self, cachetype, relpath):
        return ("-cache-%s-rel-%s.html" % (cachetype, bool(relpath))).lower()

    def __init__(self, dp, refresh=0):
        DefaultRenderer.__init__(self, dp, refresh=refresh)
        self.rendercache = {}
        self._regenerated = False

    def regenerate(self):
        """This is meant to be implemented by subclasses. Called if any cache element is out of date,
        to do the actual heavy work."""
        pass

    def checkCache(self, cachetype, relpath):
        """checks cache for rendered html code.
        'cachetype' can be any string (usually the name of a method, e.g. "Thumbnail", or "InTable")
        relpath is the relative path, here treated as boolean.
        If cache is up-to-date, returns tuple of cachekey,html (NB: cachekey is path to cache file)
        If cache is out of date, calls regenerate() if not already done so, and returns tuple of cachekey,None.
        """
        cache_ext = self._cacheFileExtension(cachetype, relpath)
        filename, path = self.subproductPath(cache_ext)
        cachekey = path
        # check already read cache
        content = self.rendercache.get(path, None)
        if content is not None:
            return content
        # if cache file is up-to-date, attempt to read content
        if self.subproductUpToDate(path):
            dprintf(3, "render cache %s is up-to-date, reading in\n", path)
            try:
                content = open(path).read()
            except:
                print(("Error reading render cache file", path, ", will regenerate"))
                traceback.print_exc()
        else:
            dprintf(3, "render cache %s is out of date, will regenerate\n", path)
        # read content? cache and return
        if content is not None:
            self.rendercache[path] = content
            return path, content
        # else regenerate
        if not self._regenerated:
            self.regenerate()
            self._regenerated = True
        # and return path,None to indicate no cache
        return path, None

    def writeCache(self, cachekey, content):
        """stores content in cache. 'cachekey' must be the same key as returned from checkCache().
        Returns the content."""
        self.rendercache[cachekey] = content
        try:
            open(cachekey, 'w').write(content)
        except:
            print(("Error writing cache file", cachekey, ", will regenerate next time"))
            traceback.print_exc()
        return content



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
#
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import math
import re
import sys
import time

from . import font
from . import pychart_util
from . import version
from .scaling import *


def _compute_bounding_box(points):
    """Given the list of coordinates (x,y), this procedure computes
    the smallest rectangle that covers all the points."""
    (xmin, ymin, xmax, ymax) = (999999, 999999, -999999, -999999)
    for p in points:
        xmin = min(xmin, p[0])
        xmax = max(xmax, p[0])
        ymin = min(ymin, p[1])
        ymax = max(ymax, p[1])
    return (xmin, ymin, xmax, ymax)


def _intersect_box(b1, b2):
    xmin = max(b1[0], b2[0])
    ymin = max(b1[1], b2[1])
    xmax = min(b1[2], b2[2])
    ymax = min(b1[3], b2[3])
    return (xmin, ymin, xmax, ymax)


def invisible_p(x, y):
    """Return true if the point (X, Y) is visible in the canvas."""
    if x < -499999 or y < -499999:
        return 1
    return 0


def to_radian(deg):
    return deg * 2 * math.pi / 360.0


def midpoint(p1, p2):
    return ((p1[0] + p2[0]) / 2.0, (p1[1] + p2[1]) / 2.0)


active_canvases = []

InvalidCoord = 999999


class T(object):
    def __init__(self):
        global active_canvases

        self.__xmax = -InvalidCoord
        self.__xmin = InvalidCoord
        self.__ymax = -InvalidCoord
        self.__ymin = InvalidCoord
        self.__clip_box = (-InvalidCoord, -InvalidCoord, InvalidCoord, InvalidCoord)
        self.__clip_stack = []
        self.__nr_gsave = 0

        self.title = theme.title or re.sub("(.*)\\.py$", "\\1", sys.argv[0])
        self.creator = theme.creator or "pychart %s" % (version.version,)
        self.creation_date = theme.creation_date or \
                             time.strftime("(%m/%d/%y) (%I:%M %p)")
        self.author = theme.author
        self.aux_comments = theme.aux_comments or ""
        active_canvases.append(self)

    def set_title(self, s):
        """Define the string to be shown in EPS/PDF "Title" field. The default value is the name of the script that creates the EPS/PDF file."""
        self.title = s

    def set_creator(self, tag):
        """Define the string to be shown in EPS %%Creator or PDF Producer field. The default value is "pychart"."""
        self.creator = tag

    def set_creation_date(self, s):
        """Define the string to be shown in EPS/PDF "CreationDate" field. Defalt value of this field is the current time."""
        self.creation_date = s

    def set_author(self, s):
        """Set the author string. Unless this method is called, the Author field is not output in EPS or PDF."""
        self.author = s

    def add_aux_comments(self, s):
        """Define an auxiliary comments to be output to the file, just after the required headers"""
        self.aux_comments += s

    def close(self):
        """This method closes the canvas and writes
        contents to the associated file.
        Calling this procedure is optional, because
        Pychart calls this procedure for every open canvas on normal exit."""
        for i in range(0, len(active_canvases)):
            if active_canvases[i] == self:
                del active_canvases[i]
                return

    def open_output(self, fname):
        """Open the output file FNAME. Returns tuple (FD, NEED_CLOSE),
        where FD is a file (or file-like) object, and NEED_CLOSE is a
        boolean flag that tells whether FD.close() should be called
        after finishing writing to the file.

        FNAME can be one of the three things:
        (1) None, in which case (sys.stdout, False) is returned.
        (2) A file-like object, in which case (fname, False) is returned.
        (3) A string, in which case this procedure opens the file and returns
        (fd, True)."""

        if not fname:
            return (sys.stdout, False)
        elif isinstance(fname, str):
            return (file(fname, "wb"), True)
        else:
            if not hasattr(fname, "write"):
                raise Exception("Expecting either a filename or a file-like object, but got %s" % fname)
            return (fname, False)

    def setbb(self, x, y):
        """Call this method when point (X,Y) is to be drawn in the
        canvas. This methods expands the bounding box to include
        this point."""
        self.__xmin = min(self.__xmin, max(x, self.__clip_box[0]))
        self.__xmax = max(self.__xmax, min(x, self.__clip_box[2]))
        self.__ymin = min(self.__ymin, max(y, self.__clip_box[1]))
        self.__ymax = max(self.__ymax, min(y, self.__clip_box[3]))

    def fill_with_pattern(self, pat, x1, y1, x2, y2):
        if invisible_p(x2, y2): return

        self.comment("FILL pat=%s (%d %d)-(%d %d)\n" % (pat, x1, y1, x2, y2))
        self.set_fill_color(pat.bgcolor)
        self._path_polygon([(x1, y1), (x1, y2), (x2, y2), (x2, y1)])
        self.fill()
        pat.draw(self, x1, y1, x2, y2)
        self.comment("end FILL.\n")

    def _path_polygon(self, points):
        "Low-level polygon-drawing routine."
        (xmin, ymin, xmax, ymax) = _compute_bounding_box(points)
        if invisible_p(xmax, ymax):
            return
        self.setbb(xmin, ymin)
        self.setbb(xmax, ymax)

        self.newpath()
        self.moveto(xscale(points[0][0]), yscale(points[0][1]))
        for point in points[1:]:
            self.lineto(xscale(point[0]), yscale(point[1]))
        self.closepath()

    def polygon(self, edge_style, pat, points, shadow=None):
        """Draw a polygon with EDGE_STYLE, fill with PAT, and the edges
        POINTS. POINTS is a sequence of coordinates, e.g., ((10,10), (15,5),
        (20,8)). SHADOW is either None or a tuple (XDELTA, YDELTA,
        fillstyle). If non-null, a shadow of FILLSTYLE is drawn beneath
        the polygon at the offset of (XDELTA, YDELTA)."""

        if pat:
            self.comment("POLYGON points=[%s] pat=[%s]"
                         % (str(points), str(pat)))
            (xmin, ymin, xmax, ymax) = _compute_bounding_box(points)

            if shadow:
                xoff, yoff, shadow_pat = shadow
                self.gsave()
                self._path_polygon(list(map(lambda p, xoff=xoff, yoff=yoff: (p[0] + xoff, p[1] + yoff), points)))
                self.clip_sub()
                self.fill_with_pattern(shadow_pat, xmin + xoff, ymin + yoff,
                                       xmax + xoff, ymax + yoff)
                self.grestore()

            self.gsave()
            self._path_polygon(points)
            self.clip_sub()
            self.fill_with_pattern(pat, xmin, ymin, xmax, ymax)
            self.grestore()
        if edge_style:
            self.comment("POLYGON points=[%s] edge=[%s]"
                         % (str(points), str(edge_style)))
            self.set_line_style(edge_style)
            self._path_polygon(points)
            self.stroke()

    def set_background(self, pat, x1, y1, x2, y2):
        xmax, xmin, ymax, ymin = self.__xmax, self.__xmin, self.__ymax, self.__ymin
        self.rectangle(None, pat, x1, y1, x2, y2)
        self.__xmax, self.__xmin, self.__ymax, self.__ymin = xmax, xmin, ymax, ymin

    def rectangle(self, edge_style, pat, x1, y1, x2, y2, shadow=None):
        """Draw a rectangle with EDGE_STYLE, fill with PAT, and the
        bounding box (X1, Y1, X2, Y2).  SHADOW is either None or a
        tuple (XDELTA, YDELTA, fillstyle). If non-null, a shadow of
        FILLSTYLE is drawn beneath the polygon at the offset of
        (XDELTA, YDELTA)."""

        self.polygon(edge_style, pat, [(x1, y1), (x1, y2), (x2, y2), (x2, y1)],
                     shadow)

    def _path_ellipsis(self, x, y, radius, ratio, start_angle, end_angle):
        self.setbb(x - radius, y - radius * ratio)
        self.setbb(x + radius, y + radius * ratio)
        oradius = nscale(radius)
        centerx, centery = xscale(x), yscale(y)
        startx, starty = centerx + oradius * math.cos(to_radian(start_angle)), \
                         centery + oradius * math.sin(to_radian(start_angle))
        self.moveto(centerx, centery)
        if start_angle % 360 != end_angle % 360:
            self.moveto(centerx, centery)
            self.lineto(startx, starty)
        else:
            self.moveto(startx, starty)
        self.path_arc(xscale(x), yscale(y), nscale(radius),
                      ratio, start_angle, end_angle)
        self.closepath()

    def ellipsis(self, line_style, pattern, x, y, radius, ratio=1.0,
                 start_angle=0, end_angle=360, shadow=None):
        """Draw an ellipsis with line_style and fill PATTERN. The center is \
        (X, Y), X radius is RADIUS, and Y radius is RADIUS*RATIO, whose \
        default value is 1.0. SHADOW is either None or a tuple (XDELTA,
        YDELTA, fillstyle). If non-null, a shadow of FILLSTYLE is drawn
        beneath the polygon at the offset of (XDELTA, YDELTA)."""

        if invisible_p(x + radius, y + radius * ratio):
            return

        if pattern:
            if shadow:
                x_off, y_off, shadow_pat = shadow
                self.gsave()
                self.newpath()
                self._path_ellipsis(x + x_off, y + y_off, radius, ratio,
                                    start_angle, end_angle)
                self.clip_sub()
                self.fill_with_pattern(shadow_pat,
                                       x - radius * 2 + x_off,
                                       y - radius * ratio * 2 + y_off,
                                       x + radius * 2 + x_off,
                                       y + radius * ratio * 2 + y_off)
                self.grestore()
            self.gsave()
            self.newpath()
            self._path_ellipsis(x, y, radius, ratio, start_angle, end_angle)
            self.clip_sub()
            self.fill_with_pattern(pattern,
                                   (x - radius * 2), (y - radius * ratio * 2),
                                   (x + radius * 2), (y + radius * ratio * 2))
            self.grestore()
        if line_style:
            self.set_line_style(line_style)
            self.newpath()
            self._path_ellipsis(x, y, radius, ratio, start_angle, end_angle)
            self.stroke()

    def clip_ellipsis(self, x, y, radius, ratio=1.0):
        """Create an elliptical clip region. You must call endclip() after
        you completed drawing. See also the ellipsis method."""

        self.gsave()
        self.newpath()
        self.moveto(xscale(x) + nscale(radius), yscale(y))
        self.path_arc(xscale(x), yscale(y), nscale(radius), ratio, 0, 360)
        self.closepath()
        self.__clip_stack.append(self.__clip_box)
        self.clip_sub()

    def clip_polygon(self, points):
        """Create a polygonal clip region. You must call endclip() after
        you completed drawing. See also the polygon method."""
        self.gsave()
        self._path_polygon(points)
        self.__clip_stack.append(self.__clip_box)
        self.__clip_box = _intersect_box(self.__clip_box, _compute_bounding_box(points))
        self.clip_sub()

    def clip(self, x1, y1, x2, y2):
        """Activate a rectangular clip region, (X1, Y1) - (X2, Y2).
        You must call endclip() after you completed drawing.

canvas.clip(x,y,x2,y2)
draw something ...
canvas.endclip()
"""

        self.__clip_stack.append(self.__clip_box)
        self.__clip_box = _intersect_box(self.__clip_box, (x1, y1, x2, y2))
        self.gsave()
        self.newpath()
        self.moveto(xscale(x1), yscale(y1))
        self.lineto(xscale(x1), yscale(y2))
        self.lineto(xscale(x2), yscale(y2))
        self.lineto(xscale(x2), yscale(y1))
        self.closepath()
        self.clip_sub()

    def endclip(self):
        """End the current clip region. When clip calls are nested, it
        ends the most recently created crip region."""
        self.__clip_box = self.__clip_stack[-1]
        del self.__clip_stack[-1]
        self.grestore()

    def curve(self, style, points):
        for p in points:
            self.setbb(p[0], p[1])
        self.newpath()
        self.set_line_style(style)
        self.moveto(xscale(points[0][0]), xscale(points[0][1]))
        i = 1
        n = 1
        while i < len(points):
            if n == 1:
                x2 = points[i]
                n += 1
            elif n == 2:
                x3 = points[i]
                n += 1
            elif n == 3:
                x4 = midpoint(x3, points[i])
                self.curveto(xscale(x2[0]), xscale(x2[1]),
                             xscale(x3[0]), xscale(x3[1]),
                             xscale(x4[0]), xscale(x4[1]))
                n = 1
            i += 1
            if n == 1:
                pass
            if n == 2:
                self.lineto(xscale(x2[0]), xscale(x2[1]))
            if n == 3:
                self.curveto(xscale(x2[0]), xscale(x2[1]),
                             xscale(x2[0]), xscale(x2[1]),
                             xscale(x3[0]), xscale(x3[1]))
            self.stroke()

    def line(self, style, x1, y1, x2, y2):
        if not style:
            return
        if invisible_p(x2, y2) and invisible_p(x1, y1):
            return

        self.setbb(x1, y1)
        self.setbb(x2, y2)

        self.newpath()
        self.set_line_style(style)
        self.moveto(xscale(x1), yscale(y1))
        self.lineto(xscale(x2), yscale(y2))
        self.stroke()

    def lines(self, style, segments):
        if not style:
            return
        (xmin, ymin, xmax, ymax) = _compute_bounding_box(segments)
        if invisible_p(xmax, ymax):
            return

        self.setbb(xmin, ymin)
        self.setbb(xmax, ymax)
        self.newpath()
        self.set_line_style(style)
        self.moveto(xscale(segments[0][0]), xscale(segments[0][1]))
        for i in range(1, len(segments)):
            self.lineto(xscale(segments[i][0]), yscale(segments[i][1]))
        self.stroke()

    def _path_round_rectangle(self, x1, y1, x2, y2, radius):
        self.moveto(xscale(x1 + radius), yscale(y1))
        self.lineto(xscale(x2 - radius), yscale(y1))
        self.path_arc(xscale(x2 - radius), yscale(y1 + radius), nscale(radius), 1, 270, 360)
        self.lineto(xscale(x2), yscale(y2 - radius))
        self.path_arc(xscale(x2 - radius), yscale(y2 - radius), nscale(radius), 1, 0, 90)
        self.lineto(xscale(x1 + radius), yscale(y2))
        self.path_arc(xscale(x1 + radius), yscale(y2 - radius), nscale(radius), 1, 90, 180)
        self.lineto(xscale(x1), xscale(y1 + radius))
        self.path_arc(xscale(x1 + radius), yscale(y1 + radius), nscale(radius), 1, 180, 270)

    def round_rectangle(self, style, fill, x1, y1, x2, y2, radius, shadow=None):
        """Draw a rectangle with rounded four corners. Parameter <radius> specifies the radius of each corner."""

        if invisible_p(x2, y2):
            return
        self.setbb(x1, y1)
        self.setbb(x2, y2)

        if fill:
            if shadow:
                x_off, y_off, shadow_fill = shadow
                self.gsave()
                self.newpath()
                self._path_round_rectangle(x1 + x_off, y1 + y_off, x2 + x_off, y2 + y_off,
                                           radius)
                self.closepath()
                self.clip_sub()
                self.fill_with_pattern(shadow_fill, x1 + x_off, y1 + y_off,
                                       x2 + x_off, y2 + y_off)
                self.grestore()

            self.gsave()
            self.newpath()
            self._path_round_rectangle(x1, y1, x2, y2, radius)
            self.closepath()
            self.clip_sub()
            self.fill_with_pattern(fill, x1, y1, x2, y2)
            self.grestore()
        if style:
            self.set_line_style(style)
            self.newpath()
            self._path_round_rectangle(x1, y1, x2, y2, radius)
            self.closepath()
            self.stroke()

    def show(self, x, y, str):
        global out
        y_org = y
        org_str = str

        if invisible_p(x, y):
            return

        (xmin, xmax, ymin, ymax) = font.get_dimension(str)

        # rectangle(line_style.default, None, x+xmin, y+ymin, x+xmax, y+ymax)
        # ellipsis(line_style.default, None, x, y, 1)
        self.setbb(x + xmin, y + ymin)
        self.setbb(x + xmax, y + ymax)

        (halign, valign, angle) = font.get_align(str)

        base_x = x
        base_y = y

        # Handle vertical alignment
        if valign == "B":
            y = font.unaligned_text_height(str)
        elif valign == "T":
            y = 0
        elif valign == "M":
            y = font.unaligned_text_height(str) / 2.0

        (xmin, xmax, ymin, ymax) = font.get_dimension(org_str)
        self.setbb(x + xmin, y_org + y + ymin)
        self.setbb(x + xmax, y_org + y + ymax)
        itr = font.text_iterator(None)

        max_width = 0

        lines = []
        for line in str.split('\n'):
            cur_width = 0
            cur_height = 0

            itr.reset(line)

            strs = []

            while 1:
                elem = next(itr)
                if not elem:
                    break

                (font_name, size, line_height, color, _h, _v, _a, str) = elem
                cur_width += font.line_width(font_name, size, str)
                max_width = max(cur_width, max_width)
                cur_height = max(cur_height, line_height)

                # replace '(' -> '\(', ')' -> '\)' to make
                # Postscript string parser happy.
                str = str.replace("(", "\\(")
                str = str.replace(")", "\\)")
                strs.append((font_name, size, color, str))
            lines.append((cur_width, cur_height, strs))

        for line in lines:
            cur_width, cur_height, strs = line
            cur_y = y - cur_height
            y = y - cur_height
            self.comment("cury: %d hei %d str %s\n" % (cur_y, cur_height, strs))
            if halign == 'C':
                cur_x = -cur_width / 2.0
            elif halign == 'R':
                cur_x = -cur_width
            else:
                cur_x = 0

            rel_x, rel_y = pychart_util.rotate(cur_x, cur_y, angle)
            self.text_begin()
            self.text_moveto(xscale(base_x + rel_x),
                             yscale(base_y + rel_y), angle)
            for segment in strs:
                font_name, size, color, str = segment
                self.text_show(font_name, nscale(size), color, str)
            self.text_end()



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import copy
import math

from . import pychart_util


def _convert_item(v, typ, line):
    if typ == "a":
        try:
            i = float(v)
        except ValueError:  # non-number
            i = v
        return i
    elif typ == "d":
        try:
            return int(v)
        except ValueError:
            raise ValueError("Can't convert %s to int; line=%s" % (v, line))
    elif typ == "f":
        try:
            return float(v)
        except ValueError:
            raise ValueError("Can't convert %s to float; line=%s" % (v, line))
    elif typ == "s":
        return v
    else:
        raise ValueError("Unknown conversion type, type=%s; line=%s" % (typ, line))


def parse_line(line, delim):
    if delim.find("%") < 0:
        return [_convert_item(item, "a", None) for item in line.split(delim)]

    data = []
    idx = 0  # indexes delim
    ch = 'f'
    sep = ','

    while idx < len(delim):
        if delim[idx] != '%':
            raise ValueError('Bad delimitor: "%s"' % delim)
        ch = delim[idx + 1]
        idx += 2
        sep = ""
        while idx < len(delim) and delim[idx] != '%':
            sep += delim[idx]
            idx += 1

        if sep != "":
            xx = line.split(sep, 1)
        else:
            # reached the end of the delimiter string.
            xx = (line,)

        data.append(_convert_item(xx[0], ch, line))
        if len(xx) >= 2:
            line = xx[1]
        else:
            line = ""
            break

    if line != "":
        for item in line.split(sep):
            data.append(_convert_item(item, ch, line))
    return data


def escape_string(str):
    return str.replace("/", "//")


def extract_rows(data, *rows):
    """Extract rows specified in the argument list.

>>> chart_data.extract_rows([[10,20], [30,40], [50,60]], 1, 2)
[[30,40],[50,60]]
"""
    try:
        # for python 2.2
        # return [data[r] for r in rows]
        out = []
        for r in rows:
            out.append(data[r])
        return out
    except IndexError:
        raise IndexError("data=%s rows=%s" % (data, rows))
    return out


def extract_columns(data, *cols):
    """Extract columns specified in the argument list.

>>> chart_data.extract_columns([[10,20], [30,40], [50,60]], 0)
[[10],[30],[50]]
"""
    out = []
    try:
        # for python 2.2:
        # return [ [r[c] for c in cols] for r in data]
        for r in data:
            col = []
            for c in cols:
                col.append(r[c])
            out.append(col)
    except IndexError:
        raise IndexError("data=%s col=%s" % (data, col))
    return out


def moving_average(data, xcol, ycol, width):
    """Compute the moving average of  YCOL'th column of each sample point
in  DATA. In particular, for each element  I in  DATA,
this function extracts up to  WIDTH*2+1 elements, consisting of
 I itself,  WIDTH elements before  I, and  WIDTH
elements after  I. It then computes the mean of the  YCOL'th
column of these elements, and it composes a two-element sample
consisting of  XCOL'th element and the mean.

>>> data = [[10,20], [20,30], [30,50], [40,70], [50,5]]
... chart_data.moving_average(data, 0, 1, 1)
[(10, 25.0), (20, 33.333333333333336), (30, 50.0), (40, 41.666666666666664), (50, 37.5)]

  The above value actually represents:

[(10, (20+30)/2), (20, (20+30+50)/3), (30, (30+50+70)/3), 
  (40, (50+70+5)/3), (50, (70+5)/2)]

"""

    out = []
    try:
        for i in range(len(data)):
            n = 0
            total = 0
            for j in range(i - width, i + width + 1):
                if j >= 0 and j < len(data):
                    total += data[j][ycol]
                    n += 1
            out.append((data[i][xcol], float(total) / n))
    except IndexError:
        raise IndexError("bad data: %s,xcol=%d,ycol=%d,width=%d" % (data, xcol, ycol, width))

    return out


def filter(func, data):
    """Parameter <func> must be a single-argument
    function that takes a sequence (i.e.,
a sample point) and returns a boolean. This procedure calls <func> on
each element in <data> and returns a list comprising elements for
which <func> returns True.

>>> data = [[1,5], [2,10], [3,13], [4,16]]
... chart_data.filter(lambda x: x[1] % 2 == 0, data)
[[2,10], [4,16]].
"""

    out = []
    for r in data:
        if func(r):
            out.append(r)
    return out


def transform(func, data):
    """Apply <func> on each element in <data> and return the list
consisting of the return values from <func>.

>>> data = [[10,20], [30,40], [50,60]]
... chart_data.transform(lambda x: [x[0], x[1]+1], data)
[[10, 21], [30, 41], [50, 61]]

"""
    out = []
    for r in data:
        out.append(func(r))
    return out


def aggregate_rows(data, col):
    out = copy.deepcopy(data)
    total = 0
    for r in out:
        total += r[col]
        r[col] = total
    return out


def empty_line_p(s):
    return s.strip() == ""


def _try_open_file(path, mode, error_message):
    try:
        fd = open(path, mode)
    except TypeError:
        fd = path
        if not getattr(fd, "readline", None):
            raise TypeError(error_message + "(got %s)" % str(fd))
    return fd


def _try_close_file(fd, path):
    if fd != path:
        fd.close()


def read_csv(path, delim=','):
    """This function reads comma-separated values from a
    file. Parameter <path> is either a pathname or a file-like object
    that supports the |readline()| method.

    Empty lines and lines
    beginning with "#" are ignored.  Parameter <delim> specifies how
    a line is separated into values. If it does not contain the
    letter "%", then <delim> marks the end of a value.
    Otherwise, this function acts like scanf in C:

chart_data.read_csv('file', '%d,%s:%d')

    Paramter <delim> currently supports
    only three conversion format specifiers:
    "d"(int), "f"(double), and "s"(string)."""

    fd = _try_open_file(path, 'r',
                        'The first argument must be a pathname or an object that supports readline() method')

    data = []
    line = fd.readline()
    while line != "":
        if line[0] != '#' and not empty_line_p(line):
            data.append(parse_line(line, delim))
        line = fd.readline()

    _try_close_file(fd, path)
    return data


def fread_csv(fd, delim=','):
    """This function is deprecated. Use read_csv instead."""
    pychart_util.warn("chart_data.fread_csv is deprecated. Use read_csv instead.")
    return read_csv(fd, delim)


def write_csv(path, data):
    """This function writes comma-separated <data> to
    <path>. Parameter <path> is either a pathname or a file-like
    object that supports the |write()| method."""

    fd = _try_open_file(path, 'w',
                        'The first argument must be a pathname or an object that supports write() method')
    for v in data:
        fd.write(",".join([str(x) for x in v]))
        fd.write("\n")
    _try_close_file(fd, path)


def fwrite_csv(fd, delim=','):
    """This function is deprecated. Use write_csv instead."""
    pychart_util.warn("chart_data.fwrite_csv is deprecated. Use write_csv instead.")
    return write_csv(fd, delim)


def read_str(delim=',', *lines):
    """This function is similar to read_csv, but it reads data from the
    list of <lines>.

fd = open("foo", "r")
data = chart_data.read_str(",", fd.readlines())"""

    data = []
    for line in lines:
        com = parse_line(line, delim)
        data.append(com)
    return data


def func(f, xmin, xmax, step=None):
    """Create sample points from function <f>, which must be a
    single-parameter function that returns a number (e.g., math.sin).
    Parameters <xmin> and <xmax> specify the first and last X values, and
    <step> specifies the sampling interval.

>>> chart_data.func(math.sin, 0, math.pi * 4, math.pi / 2)
[(0, 0.0), (1.5707963267948966, 1.0), (3.1415926535897931, 1.2246063538223773e-16), (4.7123889803846897, -1.0), (6.2831853071795862, -2.4492127076447545e-16), (7.8539816339744828, 1.0), (9.4247779607693793, 3.6738190614671318e-16), (10.995574287564276, -1.0)]

"""

    data = []
    x = xmin
    if not step:
        step = (xmax - xmin) / 100.0
    while x < xmax:
        data.append((x, f(x)))
        x += step
    return data


def _nr_data(data, col):
    nr_data = 0
    for d in data:
        nr_data += d[col]
    return nr_data


def median(data, freq_col=1):
    """Compute the median of the <freq_col>'th column of the values is <data>.

>>> chart_data.median([(10,20), (20,4), (30,5)], 0)
20
>>> chart_data.median([(10,20), (20,4), (30,5)], 1)
5.
    """

    nr_data = _nr_data(data, freq_col)
    median_idx = nr_data / 2
    i = 0
    for d in data:
        i += d[freq_col]
        if i >= median_idx:
            return d
    raise Exception("??? median ???")


def cut_extremes(data, cutoff_percentage, freq_col=1):
    nr_data = _nr_data(data, freq_col)
    min_idx = nr_data * cutoff_percentage / 100.0
    max_idx = nr_data * (100 - cutoff_percentage) / 100.0
    r = []

    i = 0
    for d in data:
        if i < min_idx:
            if i + d[freq_col] >= min_idx:
                x = copy.deepcopy(d)
                x[freq_col] = x[freq_col] - (min_idx - i)
                r.append(x)
            i += d[freq_col]
            continue
        elif i + d[freq_col] >= max_idx:
            if i < max_idx and i + d[freq_col] >= max_idx:
                x = copy.deepcopy(d)
                x[freq_col] = x[freq_col] - (max_idx - i)
                r.append(x)
            break
        i += d[freq_col]
        r.append(d)
    return r


def mean(data, val_col, freq_col):
    nr_data = 0
    sum = 0
    for d in data:
        sum += d[val_col] * d[freq_col]
        nr_data += d[freq_col]
    if nr_data == 0:
        raise IndexError("data is empty")

    return sum / float(nr_data)


def mean_samples(data, xcol, ycollist):
    """Create a sample list that contains
    the mean of the original list.

>>> chart_data.mean_samples([ [1, 10, 15], [2, 5, 10], [3, 8, 33] ], 0, (1, 2))
[(1, 12.5), (2, 7.5), (3, 20.5)]
"""
    out = []
    numcol = len(ycollist)
    try:
        for elem in data:
            v = 0
            for col in ycollist:
                v += elem[col]
            out.append((elem[xcol], float(v) / numcol))
    except IndexError:
        raise IndexError("bad data: %s,xcol=%d,ycollist=%s" % (data, xcol, ycollist))

    return out


def stddev_samples(data, xcol, ycollist, delta=1.0):
    """Create a sample list that contains the mean and standard deviation of the original list. Each element in the returned list contains following values: [MEAN, STDDEV, MEAN - STDDEV*delta, MEAN + STDDEV*delta].

>>> chart_data.stddev_samples([ [1, 10, 15, 12, 15], [2, 5, 10, 5, 10], [3, 32, 33, 35, 36], [4,16,66, 67, 68] ], 0, range(1,5))
[(1, 13.0, 2.1213203435596424, 10.878679656440358, 15.121320343559642), (2, 7.5, 2.5, 5.0, 10.0), (3, 34.0, 1.5811388300841898, 32.418861169915807, 35.581138830084193), (4, 54.25, 22.094965489902897, 32.155034510097103, 76.344965489902904)]
"""
    out = []
    numcol = len(ycollist)
    try:
        for elem in data:
            total = 0
            for col in ycollist:
                total += elem[col]
            mean = float(total) / numcol
            variance = 0
            for col in ycollist:
                variance += (mean - elem[col]) ** 2
            stddev = math.sqrt(variance / numcol) * delta
            out.append((elem[xcol], mean, stddev, mean - stddev, mean + stddev))



    except IndexError:
        raise IndexError("bad data: %s,xcol=%d,ycollist=%s" % (data, xcol, ycollist))
    return out


def nearest_match(data, col, val):
    min_delta = None
    match = None

    for d in data:
        if min_delta == None or abs(d[col] - val) < min_delta:
            min_delta = abs(d[col] - val)
            match = d
    pychart_util.warn("XXX ", match)
    return match



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import re
import string

from . import color
from . import pychart_util
from . import theme
from .afm import dir as dir_

__doc__ = """The module for manipulating texts and their attributes.

Pychart supports extensive sets of attributes in texts. All attributes
are specified via "escape sequences", starting from letter "/". For
example, the below examples draws string "Hello" using a 12-point font
at 60-degree angle:

/12/a60{}Hello

List of attributes:

/hA
    Specifies horizontal alignment of the text.  A is one of L (left
    alignment), R (right alignment), or C (center alignment).
/vA
    Specifies vertical alignment of the text.  A is one of "B"
    (bottom), "T" (top), " M" (middle).

/F{FONT}
    Switch to FONT font family.
/T
    Shorthand of /F{Times-Roman}.
/H
    Shorthand of /F{Helvetica}.
/C
    Shorthand of /F{Courier}.
/B
    Shorthand of /F{Bookman-Demi}.
/A
    Shorthand of /F{AvantGarde-Book}.
/P
    Shorthand of /F{Palatino}.
/S
    Shorthand of /F{Symbol}.
/b
    Switch to bold typeface.
/i
    Switch to italic typeface.
/o
    Switch to oblique typeface.
/DD
    Set font size to DD points.

    /20{}2001 space odyssey!

/cDD
    Set gray-scale to 0.DD. Gray-scale of 00 means black, 99 means white.

//, /{, /}
    Display `/', `{', or `}'.
    
{ ... }
    Limit the effect of escape sequences. For example, the below
    example draws "Foo" at 12pt, "Bar" at 8pt, and "Baz" at 12pt.

    /12Foo{/8Bar}Baz
\n
    Break the line.
"""

# List of fonts for which their absence have been already warned.
_undefined_font_warned = {}  # type: dict


class FontException(Exception):
    def __init__(self, msg, str):
        self.__msg = msg
        self.__str = str

    def __str__(self):
        return """%s (got "%s"). Write "//". "/{", "/}" to display "/", "{", "}", respectively.""" % \
               (self.__msg, self.__str)


def _intern_afm(font, text):
    global _undefined_font_warned

    r = dir_.afm.get(font, None)
    if r: return r

    font2 = _font_aliases.get(font, None)
    if font2:
        r = dir_.afm.get(font2, None)
        if r: return r

    try:
        exec ("import pychart.afm.%s" % re.sub("-", "_", font))
        return dir_.afm[font]
    except:
        if not font2 and font not in _undefined_font_warned:
            pychart_util.warn('Warning: unknown font "%s" while parsing "%s"'
                              % (font, text))
            _undefined_font_warned[font] = 1

    if font2:
        try:
            exec ('import pychart.afm.%s' % re.sub('-', '_', font2))
            return dir_.afm[font2]
        except:
            if font not in _undefined_font_warned:
                pychart_util.warn('Warning: unknown font "%s" while parsing "%s"' % (font, text))
                _undefined_font_warned[font] = 1
    return None


def line_width(font, size, text):
    table = _intern_afm(font, text)
    if not table:
        return 0

    width = 0
    for ch in text:
        code = ord(ch)
        if code < len(table):
            width += table[code]
        else:
            # Invalid char. Make up a number.
            width += table[0]

    width = float(width) * size / 1000.0
    return width


_font_family_map = {'T': 'Times',
                    'H': 'Helvetica',
                    'C': 'Courier',
                    'N': 'Helvetica-Narrow',
                    'B': 'Bookman-Demi',
                    'A': 'AvantGarde-Book',
                    'P': 'Palatino',
                    'S': 'Symbol'}

# Aliases for ghostscript font names.

_font_aliases = {
    'Bookman-Demi': 'URWBookmanL-DemiBold%I',
    'Bookman-DemiItalic': 'URWBookmanL-DemiBoldItal',
    'Bookman-Demi-Italic': 'URWBookmanL-DemiBoldItal',
    'Bookman-Light': 'URWBookmanL-Ligh',
    'Bookman-LightItalic': 'URWBookmanL-LighItal',
    'Bookman-Light-Italic': 'URWBookmanL-LighItal',
    'Courier': 'NimbusMonL-Regu',
    'Courier-Oblique': 'NimbusMonL-ReguObli',
    'Courier-Bold': 'NimbusMonL-Bold',
    'Courier-BoldOblique': 'NimbusMonL-BoldObli',
    'AvantGarde-Book': 'URWGothicL-Book',
    'AvantGarde-BookOblique': 'URWGothicL-BookObli',
    'AvantGarde-Book-Oblique': 'URWGothicL-BookObli',
    'AvantGarde-Demi': 'URWGothicL-Demi',
    'AvantGarde-DemiOblique': 'URWGothicL-DemiObli',
    'AvantGarde-Demi-Oblique': 'URWGothicL-DemiObli',
    'Helvetica': 'NimbusSanL-Regu',
    'Helvetica-Oblique': 'NimbusSanL-ReguItal',
    'Helvetica-Bold': 'NimbusSanL-Bold',
    'Helvetica-BoldOblique': 'NimbusSanL-BoldItal',
    'Helvetica-Narrow': 'NimbusSanL-ReguCond',
    'Helvetica-Narrow-Oblique': 'NimbusSanL-ReguCondItal',
    'Helvetica-Narrow-Bold': 'NimbusSanL-BoldCond',
    'Helvetica-Narrow-BoldOblique': 'NimbusSanL-BoldCondItal',
    'Palatino-Roman': 'URWPalladioL-Roma',
    'Palatino': 'URWPalladioL-Roma',
    'Palatino-Italic': 'URWPalladioL-Ital',
    'Palatino-Bold': 'URWPalladioL-Bold',
    'Palatino-BoldItalic': 'URWPalladioL-BoldItal',
    'NewCenturySchlbk-Roman': 'CenturySchL-Roma',
    'NewCenturySchlbk': 'CenturySchL-Roma',
    'NewCenturySchlbk-Italic': 'CenturySchL-Ital',
    'NewCenturySchlbk-Bold': 'CenturySchL-Bold',
    'NewCenturySchlbk-BoldItalic': 'CenturySchL-BoldItal',
    'Times-Roman': 'NimbusRomNo9L-Regu',
    'Times': 'NimbusRomNo9L-Regu',
    'Times-Italic': 'NimbusRomNo9L-ReguItal',
    'Times-Bold': 'NimbusRomNo9L-Medi',
    'Times-BoldItalic': 'NimbusRomNo9L-MediItal',
    'Symbol': 'StandardSymL',
    'ZapfChancery-MediumItalic': 'URWChanceryL-MediItal',
    'ZapfChancery-Medium-Italic': 'URWChanceryL-MediItal',
    'ZapfDingbats': 'Dingbats'
}


class text_state:
    def copy(self):
        ts = text_state()
        ts.family = self.family
        ts.modifiers = list(self.modifiers)
        ts.size = self.size
        ts.line_height = self.line_height
        ts.color = self.color
        ts.halign = self.halign
        ts.valign = self.valign
        ts.angle = self.angle
        return ts

    def __init__(self):
        self.family = theme.default_font_family
        self.modifiers = []  # 'b' for bold, 'i' for italic, 'o' for oblique.
        self.size = theme.default_font_size
        self.line_height = theme.default_line_height or theme.default_font_size
        self.color = color.default
        self.halign = theme.default_font_halign
        self.valign = theme.default_font_valign
        self.angle = theme.default_font_angle


class text_iterator:
    def __init__(self, s):
        self.str = str(s)
        self.i = 0
        self.ts = text_state()
        self.stack = []

    def reset(self, s):
        self.str = str(s)
        self.i = 0

    def __return_state(self, ts, str):
        font_name = ts.family

        if ts.modifiers != []:
            is_bold = 0
            if 'b' in ts.modifiers:
                is_bold = 1
                font_name += '-Bold'
            if 'o' in ts.modifiers:
                if not is_bold:
                    font_name += '-'
                font_name += 'Oblique'
            elif 'i' in ts.modifiers:
                if not is_bold:
                    font_name += '-'
                font_name += 'Italic'
        elif font_name in ('Palatino', 'Times', 'NewCenturySchlbk'):
            font_name += '-Roman'

        return (font_name, ts.size, ts.line_height, ts.color,
                ts.halign, ts.valign, ts.angle, str)

    def __parse_float(self):
        istart = self.i
        while self.i < len(self.str) and self.str[self.i] in string.digits or self.str[self.i] == '.':
            self.i += 1
        return float(self.str[istart:self.i])

    def __parse_int(self):
        istart = self.i
        while self.i < len(self.str) and \
                (self.str[self.i] in string.digits or
                 self.str[self.i] == '-'):
            self.i += 1
        return int(self.str[istart:self.i])

    def __next__(self):
        "Get the next text segment. Return an 8-element array: (FONTNAME, SIZE, LINEHEIGHT, COLOR, H_ALIGN, V_ALIGN, ANGLE, STR."
        l = []
        changed = 0
        self.old_state = self.ts.copy()

        while self.i < len(self.str):
            if self.str[self.i] == '/':
                self.i = self.i + 1
                ch = self.str[self.i]
                self.i = self.i + 1
                self.old_state = self.ts.copy()
                if ch == '/' or ch == '{' or ch == '}':
                    l.append(ch)
                elif ch in _font_family_map:
                    self.ts.family = _font_family_map[ch]
                    changed = 1
                elif ch == 'F':
                    # /F{font-family}
                    if self.str[self.i] != '{':
                        raise FontException('"{" must follow /F', self.str)
                    self.i += 1
                    istart = self.i
                    while self.str[self.i] != '}':
                        self.i += 1
                        if self.i >= len(self.str):
                            raise FontException('Expecting "/F{...}"', self.str)
                    self.ts.family = self.str[istart:self.i]
                    self.i += 1
                    changed = 1

                elif ch in string.digits:
                    self.i -= 1
                    self.ts.size = self.__parse_int()
                    self.ts.line_height = self.ts.size
                    changed = 1
                elif ch == 'l':
                    self.ts.line_height = self.__parse_int()
                    changed = 1
                elif ch == 'b':
                    self.ts.modifiers.append('b')
                    changed = 1
                elif ch == 'i':
                    self.ts.modifiers.append('i')
                    changed = 1
                elif ch == 'o':
                    self.ts.modifiers.append('q')
                    changed = 1
                elif ch == 'c':
                    self.ts.color = color.gray_scale(self.__parse_float())
                elif ch == 'v':
                    if self.str[self.i] not in 'BTM':
                        raise FontException('Undefined escape sequence "/v%c"' %
                                            self.str[self.i], self.str)
                    self.ts.valign = self.str[self.i]
                    self.i += 1
                    changed = 1
                elif ch == 'h':
                    if self.str[self.i] not in 'LRC':
                        raise FontException('Undefined escape sequence "/h%c"' %
                                            self.str[self.i], self.str)
                    self.ts.halign = self.str[self.i]
                    self.i += 1
                    changed = 1
                elif ch == 'a':
                    self.ts.angle = self.__parse_int()
                    changed = 1
                else:
                    raise FontException('Undefined escape sequence: "/%c"' % ch,
                                        self.str)
            elif self.str[self.i] == '{':
                self.stack.append(self.ts.copy())
                self.i += 1
            elif self.str[self.i] == '}':
                if len(self.stack) == 0:
                    raise FontError('Unmatched "}"', self.str)
                self.ts = self.stack[-1]
                del self.stack[-1]
                self.i += 1
                changed = 1
            else:
                l.append(self.str[self.i])
                self.i += 1

            if changed and len(l) > 0:
                return self.__return_state(self.old_state, ''.join(l))
            else:
                # font change in the beginning of the sequence doesn't count.
                self.old_state = self.ts.copy()
                changed = 0
        if len(l) > 0:
            return self.__return_state(self.old_state, ''.join(l))
        else:
            return None


#
#

def unaligned_get_dimension(text):
    """Return the bounding box of the text, assuming that the left-bottom corner
    of the first letter of the text is at (0, 0). This procedure ignores
    /h, /v, and /a directives when calculating the BB; it just returns the
    alignment specifiers as a part of the return value. The return value is a
    tuple (width, height, halign, valign, angle)."""

    xmax = 0
    ymax = 0
    ymax = 0
    angle = None
    halign = None
    valign = None
    itr = text_iterator(None)
    for line in str(text).split('\n'):
        cur_height = 0
        cur_width = 0
        itr.reset(line)
        while 1:
            elem = next(itr)
            if not elem:
                break
            (font, size, line_height, color, new_h, new_v, new_a, chunk) = elem
            if halign != None and new_h != halign:
                raise FontException('Only one "/h" can appear in a string.',
                                    str(text))
            if valign != None and new_v != valign:
                raise FontException('Only one "/v" can appear in a string.',
                                    str(text))
            if angle != None and new_a != angle:
                raise FontException('Only one "/a" can appear in a string.',
                                    str(text))
            halign = new_h
            valign = new_v
            angle = new_a
            cur_width += line_width(font, size, chunk)
            cur_height = max(cur_height, line_height)
        xmax = max(cur_width, xmax)
        ymax += cur_height
    return (xmax, ymax,
            halign or theme.default_font_halign,
            valign or theme.default_font_valign,
            angle or theme.default_font_angle)


def get_dimension(text):
    """Return the bounding box of the <text>,
    assuming that the left-bottom corner
    of the first letter of the text is at (0, 0). This procedure ignores
    /h, /v, and /a directives when calculating the boundingbox; it just returns the
    alignment specifiers as a part of the return value. The return value is a
    tuple (width, height, halign, valign, angle)."""
    (xmax, ymax, halign, valign, angle) = unaligned_get_dimension(text)
    xmin = ymin = 0
    if halign == 'C':
        xmin = -xmax / 2.0
        xmax = xmax / 2.0
    elif halign == 'R':
        xmin = -xmax
        xmax = 0
    if valign == 'M':
        ymin = -ymax / 2.0
        ymax = ymax / 2.0
    elif valign == 'T':
        ymin = -ymax
        ymax = 0
    if angle != 0:
        (x0, y0) = pychart_util.rotate(xmin, ymin, angle)
        (x1, y1) = pychart_util.rotate(xmax, ymin, angle)
        (x2, y2) = pychart_util.rotate(xmin, ymax, angle)
        (x3, y3) = pychart_util.rotate(xmax, ymax, angle)
        xmax = max(x0, x1, x2, x3)
        xmin = min(x0, x1, x2, x3)
        ymax = max(y0, y1, y2, y3)
        ymin = min(y0, y1, y2, y3)
        return (xmin, xmax, ymin, ymax)
    return (xmin, xmax, ymin, ymax)


def unaligned_text_width(text):
    x = unaligned_get_dimension(text)
    return x[0]


def text_width(text):
    """Return the width of the <text> in points."""
    (xmin, xmax, d1, d2) = get_dimension(text)
    return xmax - xmin


def unaligned_text_height(text):
    x = unaligned_get_dimension(text)
    return x[1]


def text_height(text):
    """Return the total height of the <text> and the length from the
    base point to the top of the text box."""
    (d1, d2, ymin, ymax) = get_dimension(text)
    return (ymax - ymin, ymax)


def get_align(text):
    "Return (halign, valign, angle) of the <text>."
    (x1, x2, h, v, a) = unaligned_get_dimension(text)
    return (h, v, a)


def quotemeta(text):
    """Quote letters with special meanings in pychart so that <text> will display
    as-is when passed to canvas.show(). 

>>> font.quotemeta("foo/bar")
"foo//bar"
"""
    text = re.sub(r'/', '//', text)
    text = re.sub(r'\\{', '/{', text)
    text = re.sub(r'\\}', '/}', text)
    return text



# -*- coding: utf-8 -*-
import os
import os.path
import traceback

try:
    from PyQt4.Qt import *
except ImportError as e:
    print("Failed to load Qt4. Purr is not available")
    raise e

import Purr
import Purr.Render
from Purr.Render import quote_url

FULLINDEX = "fullindex.html"
INDEX = "index.html"

PURR_ICONS = {16: "purr16.png", 24: "purr24.png", 32: "purr32.png"}


def initIndexDir(logdir):
    for sz, filename in list(PURR_ICONS.items()):
        path = os.path.join(logdir, filename)
        if not os.path.exists(path):
            Purr.pixmaps.purr_logo.pm().scaled(QSize(sz, sz), Qt.KeepAspectRatioByExpanding,
                                               Qt.SmoothTransformation).save(path)


def renderIcon(size, path=None):
    filename = PURR_ICONS[size]
    if path:
        filename = os.path.join(path, filename)
    return """<IMG SRC="%s"></IMG>""" % quote_url(filename)


def writeLogIndex(logdir, title, timestamp, entries, refresh=0):
    logdir = os.path.normpath(os.path.abspath(logdir))
    fullindex = os.path.join(logdir, FULLINDEX)
    tocindex = os.path.join(logdir, INDEX)

    # if 5 entries or less, do an index only, and remove the fullindex
    if len(entries) <= 5:
        if os.path.exists(fullindex):
            try:
                os.remove(fullindex)
            except:
                print(("Error removing %s:" % fullindex))
                traceback.print_exc()
        fullindex = tocindex
        tocindex = None

    # generate TOC index, if asked to
    if tocindex:
        fobj = open(tocindex, "wt")
        # write index.html
        fobj.write("""<HTML><BODY>\n
      <TITLE>%s</TITLE>

      <H1><A CLASS="TITLE" TIMESTAMP=%.6f>%s</A></H1>

      """ % (title, timestamp, title))
        fobj.write(
            """<DIV ALIGN=right><P><A HREF=%s>Printable version (single HTML page).</A></P></DIV>\n\n""" % FULLINDEX)
        # write entries
        for i, entry in enumerate(entries):
            path = os.path.normpath(os.path.abspath(entry.index_file))
            if path.startswith(logdir + '/'):
                path = path[(len(logdir) + 1):]
            fobj.write("""<P><A HREF="%s">%d. %s</A></P>\n\n""" % (quote_url(path), i + 1, entry.title))
        # write footer
        fobj.write("<HR>\n\n")
        fobj.write("""<DIV ALIGN=right>%s <I><SMALL>This log was generated
               by PURR version %s.</SMALL></I></DIV>\n""" % (renderIcon(16), Purr.Version))
        fobj.write("</BODY></HTML>\n")
        fobj = None

    # generate full index, if asked to
    if fullindex:
        fobj = open(fullindex, "wt")
        # write index.html
        fobj.write("""<HTML><BODY>\n
      <TITLE>%s</TITLE>

      <H1><A CLASS="TITLE" TIMESTAMP=%.6f>%s</A></H1>

      """ % (title, timestamp, title))
        # write entries
        for entry in entries:
            fobj.write(entry.renderIndex(os.path.join(os.path.basename(entry.pathname), ""), refresh=refresh))
        # write footer
        fobj.write("<HR>\n")
        fobj.write("""<DIV ALIGN=right>%s <I><SMALL>This log was generated
               by PURR version %s.</SMALL></I></DIV>\n""" % (renderIcon(16), Purr.Version))
        fobj.write("</BODY></HTML>\n")
        fobj = None



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
#
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#

import math

from . import basecanvas
from . import pychart_util
from .scaling import *

try:
    import zlib

    _zlib_available_p = 1
except:
    _zlib_available_p = 0


class pdf_stream(object):
    def __init__(self, fp):
        self.fp = fp
        self.off = 0

    def write(self, str):
        self.fp.write(str)
        self.off += len(str)

    def tell(self):
        return self.off


def to_radian(deg):
    return deg * 2 * math.pi / 360.0


class T(basecanvas.T):
    def __init__(self, fname, compress_p_):
        basecanvas.T.__init__(self)
        self.__out_fname = fname
        self.__reset_context()
        self.__next_obj_id = 1
        self.__next_font_id = 1
        self.__obj_offsets = {}

        # Maps font name -> integer font ID.
        self.__registered_fonts = {}

        self.__lines = []
        self.__nr_gsave = 0

        if compress_p_ and not _zlib_available_p:
            pychart_util.warn("Zlib not available. Compression request ignored.\n")
            compress_p_ = 0
        self.__compress_p = compress_p_

    def __register_font(self, name):
        "Assign an ID to the font NAME. Return its ID."
        if name not in self.__registered_fonts:
            self.__registered_fonts[name] = self.__next_font_id
            self.__next_font_id += 1
        return self.__registered_fonts[name]

    def __define_obj(self, fp, str):
        obj_id = self.__next_obj_id
        self.__next_obj_id += 1
        self.__obj_offsets[obj_id] = fp.tell()
        fp.write("%d 0 obj\n%s\nendobj\n" % (obj_id, str))
        return obj_id

    def __define_stream_obj(self, fp, s):
        if self.__compress_p:
            p = zlib.compress(s)
            return self.__define_obj(fp, "<</Length %d/Filter/FlateDecode>>\nstream\n%sendstream"
                                     % (len(p), p))
        else:
            return self.__define_obj(fp, "<</Length %d\n>>\nstream\n%s\nendstream"
                                     % (len(s), s))

    def __define_font_obj(self, fp, name, font_id):
        obj_id = self.__define_obj(fp,
                                   """<</Type/Font /Subtype/Type1 /Name/F%d /BaseFont/%s /Encoding/MacRomanEncoding>>""" % (
                                   font_id, name))
        return obj_id

    def __reset_context(self):
        self.__font_name = None
        self.__font_size = -1
        self.__line_style = None
        self.__fill_color = None
        self.__stroke_color = None
        self.__mtx_pushed = 0

    def newpath(self):
        pass

    def set_fill_color(self, color):
        if self.__fill_color == color:
            return
        if color.r == color.g and color.r == color.b:
            self.__write("%f g\n" % (color.r))
            self.__write("%f G\n" % (color.r))
        else:
            self.__write("%f %f %f rg\n" % (color.r, color.g, color.b))
            self.__write("%f %f %f RG\n" % (color.r, color.g, color.b))
        self.__fill_color = color

    def set_stroke_color(self, color):
        self.set_fill_color(color)
        return

    def __arcsub(self, x, y, radius, start, theta):
        xcos = math.cos(to_radian(theta))
        xsin = math.sin(to_radian(theta))
        x0 = radius * xcos
        y0 = radius * xsin
        x1 = radius * (4 - xcos) / 3.0
        y1 = radius * (1 - xcos) * (xcos - 3) / (3 * xsin)

        xx0, xy0 = pychart_util.rotate(x0, y0, start + theta)
        xx1, xy1 = pychart_util.rotate(x1, -y1, start + theta)
        xx2, xy2 = pychart_util.rotate(x1, y1, start + theta)
        self.__write("%f %f %f %f %f %f c\n" %
                     (x + xx1, y + xy1, x + xx2, y + xy2, x + xx0, y + xy0))

    def path_arc(self, x, y, radius, ratio, start, end):
        self.comment("PATHARC %f %f %f %f %f %f\n"
                     % (x, y, radius, ratio, start, end))
        step = 10
        if radius < 10:
            step = 20
        if radius < 5:
            step = 30
        if ratio != 1.0:
            self.push_transformation((x, y), (1, ratio), None)
            deg = start
            while deg < end:
                theta = min(step, end - deg)
                self.__arcsub(x, y, radius, deg, theta / 2)
                deg += theta
            self.pop_transformation()
        else:
            deg = start
            while deg < end:
                theta = min(step, end - deg)
                self.__arcsub(x, y, radius, deg, theta / 2)
                deg += theta
        self.comment("end PATHARC\n")

    def text_begin(self):
        self.__write("BT ")
        self.__font_name = None
        self.__font_size = None

    def text_end(self):
        self.__write("ET\n")

    def text_moveto(self, x, y, angle):
        if angle != None:
            xcos = math.cos(to_radian(angle))
            xsin = math.sin(to_radian(angle))
            self.__write("%f %f %f %f %f %f Tm " % (xcos, xsin, -xsin, xcos, x, y))
        else:
            self.__write("1 0 0 1 %f %f Tm " % (x, y))

    def text_show(self, font_name, font_size, color, str):
        if self.__font_name != font_name or self.__font_size != font_size:
            self.__write("/F%d %d Tf " % (self.__register_font(font_name), font_size))
            self.__font_name = font_name
            self.__font_size = font_size
        self.set_fill_color(color)
        self.__write("(%s) Tj " % (str.encode('mac_roman')))

    def push_transformation(self, baseloc, scale, angle, in_text=0):
        if in_text:
            op = "Tm"
        else:
            op = "cm"
            self.gsave()

        if baseloc == None:
            baseloc = (0, 0)

        if angle != None:
            radian = to_radian(angle)
            self.__write("%f %f %f %f %f %f %s\n" %
                         (math.cos(radian), math.sin(radian),
                          -math.sin(radian), math.cos(radian),
                          baseloc[0], baseloc[1], op))
        if scale != None:
            self.__write("%f 0 0 %f %f %f %s\n" % (scale[0], scale[1],
                                                   baseloc[0],
                                                   baseloc[1], op))

    def pop_transformation(self, in_text=0):
        if not in_text:
            self.grestore()

    def closepath(self):
        self.__write("h\n")

    def clip_sub(self):
        self.__write("W n\n")

    def fill(self):
        self.__write("f n\n")

    def gsave(self):
        self.__write("q\n")

    def grestore(self):
        self.__write("Q\n")
        self.__reset_context()

    def moveto(self, x, y):
        self.__write('%f %f m ' % (x, y))

    def lineto(self, x, y):
        self.__write("%f %f l\n" % (x, y))

    def stroke(self):
        self.__write("S\n")

    def set_line_style(self, style):
        if (self.__line_style == style):
            pass
        else:
            self.set_stroke_color(style.color)
            self.__write("%f w " % nscale(style.width))
            if style.dash != None:
                self.__write("[%s] 0 d\n" %
                             " ".join(map(str, nscale_seq(style.dash))))
            else:
                self.__write("[] 0 d\n")
            self.__write("%d j %d J\n" % (style.cap_style, style.join_style))
            self.__line_style = style

    def comment(self, str):
        if not self.__compress_p:
            self.__write("%%" + str + "\n")

    def verbatim(self, str):
        self.__write(str)

    def __write(self, str):
        self.__lines.append(str)

    def close(self):
        basecanvas.T.close(self)
        if self.__lines == []:
            return

        _fp, need_close = self.open_output(self.__out_fname)
        fp = pdf_stream(_fp)

        fp.write("%PDF-1.2\n")

        stream_obj_id = self.__define_stream_obj(fp, " ".join(self.__lines))

        fontstr = ""
        for font_name, font_id in list(self.__registered_fonts.items()):
            obj_id = self.__define_font_obj(fp, font_name, font_id)
            fontstr += "/F%d %d 0 R " % (font_id, obj_id)

        pages_obj_id = self.__define_obj(fp, " <</Type/Pages /Kids [%d 0 R] /Count 1 >>" % (self.__next_obj_id + 1))

        bbox = theme.adjust_bounding_box([xscale(self.__xmin), yscale(self.__ymin),
                                          xscale(self.__xmax), yscale(self.__ymax)])

        page_obj_id = self.__define_obj(fp, """  <</Type/Page
\t/Parent %d 0 R
\t/Contents %d 0 R
\t/MediaBox [%d %d %d %d]
\t/Resources << /ProcSet [/PDF /Text]
\t\t/Font << %s >>
>> >>""" % (pages_obj_id, stream_obj_id,
            bbox[0], bbox[1], bbox[2], bbox[3], fontstr))

        info_str = "/Producer (%s)\n/CreationDate (%s)" % (self.creator, self.creation_date)

        if self.title:
            info_str += "\n/Title (%s)" % (self.title,)
        if self.author:
            info_str += "\n/Author (%s)" % (self.author,)

        info_obj_id = self.__define_obj(fp, """<<%s>>""" % info_str)
        catalog_obj_id = self.__define_obj(fp, """  <</Type/Catalog/Pages %d 0 R>>""" % (pages_obj_id))

        xref_offset = fp.tell()
        fp.write("xref\n0 %d\n" % (len(self.__obj_offsets) + 1))
        fp.write("0000000000 65535 f \n")
        id = 1
        while id <= len(self.__obj_offsets):
            fp.write("%010d 00000 n \n" % (self.__obj_offsets[id]))
            id += 1
        fp.write("trailer << /Size %d /Root %d 0 R /Info %d 0 R\n>>\n" % (
        len(self.__obj_offsets) + 1, catalog_obj_id, info_obj_id))
        fp.write("startxref\n%d\n%%%%EOF\n" % xref_offset)

        if need_close:
            _fp.close()



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import canvas
from . import coord


class T(coord.T):
    def __init__(self, data, col):

        """This attribute is meaningful only when x_coord_system ==
        'category'. This attribute selects the column of
        'x_category_data' from which X values are computed.
        Meaningful only when x_coord_system == 'category'.  This
        attribute specifies the data-set from which the X values are
        extracted. See also x_category_col."""

        self.data = data
        self.col = col

    def get_canvas_pos(self, size, val, min, max):
        i = 0.5
        for v in self.data:
            if v[self.col] == val:
                return size * i / float(len(self.data))
            i += 1
        # the drawing area is clipped. So negative offset will make this plot
        # invisible.
        return canvas.invalid_coord

    def get_tics(self, min, max, interval):
        tics = []
        if interval == None: interval = 1

        for i in range(0, len(self.data), interval):
            tics.append(self.data[i][self.col])
        return tics
        # return map(lambda pair, self = self: pair[self.col], self.data)



"""Purr.Pipe is a very lightweight module for writing to Purr pipes.
It is meant to be used within TDL scripts and such."""

import os.path
import traceback

_pipefile = ".purr.pipe"


class Pipe(object):
    """The Pipe class represents a Purr pipe."""

    def __init__(self, dirname):
        """opens a Purr pipe to the specified directory"""
        self.dirname = dirname
        global _pipefile
        file = os.path.join(dirname, _pipefile)
        self.pipefile = os.path.abspath(os.path.normpath(os.path.realpath(file)))
        self.pipefile_read = os.path.abspath(os.path.normpath(os.path.realpath(file))) + ".read"
        self.mtime = 0

    # writer section

    def _write(self, what):
        """writes something to the Purr pipe"""
        try:
            open(self.pipefile, "a").write(what)
        except:
            print(("Error writing to %s:" % self.pipefile))
            traceback.print_exc()

    def title(self, title, show=False):
        """writes a title tag to the Purr pipe"""
        self._write("title:%d:%s\n" % (int(show), title))
        return self

    def comment(self, comment, show=False, endline=True, newpar=False):
        """writes a comment tag to the Purr pipe"""
        if not endline:
            comment += "<NOBR>"
        if newpar:
            comment += "<BR>"
        self._write("comment:%d:%s\n" % (int(show), comment))
        return self

    def __lshift__(self, comment):
        """<< is equivalent to writing a comment"""
        return self.comment(comment)

    def pounce(self, file, show=False):
        """writes a pounce command to the Purr pipe"""
        file = os.path.abspath(os.path.normpath(os.path.realpath(file)))
        self._write("pounce:%d:%s\n" % (int(show), file))
        return self

    # reader section
    def read(self):
        """Checks if anything has arrived in the pipe file, returns list of
        (command,show,content) tuples.
        Special tags added in comment() above are replaced.
        """
        # read file is set to None on error, so that we stop reading
        if self.pipefile_read is None:
            return []
        # no pipe -- read nothing
        if not os.path.exists(self.pipefile):
            return []
        # move pipe file out of the way before we start reading from it
        try:
            os.rename(self.pipefile, self.pipefile_read)
        except:
            print(("Error renaming %s to %s:" % self.pipefile, self.pipefile_read))
            traceback.print_exc()
            self.pipefile_read = None
            return []
        # now read everything
        try:
            lines = open(self.pipefile_read).readlines()
        except:
            print(("Error reading %s:" % self.pipefile_read))
            traceback.print_exc()
            self.pipefile_read = None
            return []
        # parse lines
        cmds = []
        for line in lines:
            tokens = line.rstrip().split(":", 2)
            if len(tokens) < 3:
                continue
            command, show, content = tokens
            # convert 'show' to bool
            try:
                show = bool(int(show))
            except:
                show = False
            # if 'comment' command, parse content string
            if command == "comment":
                # <BR> inserts new paragraph (double-newline)
                content = content.replace("<BR>", "\n\n")
                # <NOBR> blocks newline at end, which is otherwise implictly added...
                if content.endswith("<NOBR>"):
                    content = content[:-6]
                # ...here
                elif not content.endswith("\n"):
                    content += "\n"
            # add to list of contents
            # print "read %s:%d:%s:"%(command,show,content)
            cmds.append((command, show, content))
        return cmds


def open(dirname):
    return Pipe(dirname)



__all__ = ["Courier_Oblique", "AvantGarde_BookOblique", "Times_Italic", "Helvetica_Bold", "NewCenturySchlbk_Roman",
           "Helvetica", "Helvetica_Narrow", "AvantGarde_Demi", "Times_BoldItalic", "Helvetica_Narrow_Bold",
           "Helvetica_Light", "Bookman_DemiItalic", "Utopia_Regular", "Times_Roman", "Palatino_Italic", "Courier_Bold",
           "ZapfChancery_MediumItalic", "NewCenturySchlbk_Italic", "NewCenturySchlbk_BoldItalic",
           "Helvetica_Narrow_BoldOblique", "Courier", "AvantGarde_DemiOblique", "Courier_BoldOblique",
           "Bookman_LightItalic", "Symbol", "Utopia_Bold", "Times_Bold", "Helvetica_BoldOblique", "Utopia_BoldItalic",
           "AvantGarde_Book", "Bookman_Demi", "Palatino_Roman", "Bookman_Light", "Utopia_Italic",
           "NewCenturySchlbk_Bold", "Helvetica_LightOblique", "ZapfDingbats", "Helvetica_Narrow_Oblique",
           "Helvetica_Oblique", "Palatino_BoldItalic", "Palatino_Bold", "dir"]



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import math

from . import coord
from . import pychart_util


class T(coord.T):
    def get_canvas_pos(self, size, val, min, max):
        return size * (val - min) / float(max - min)

    def get_tics(self, min, max, interval):
        v = []
        x = min
        while x <= max:
            v.append(x)
            x += interval
        return v

    def get_min_max(self, dmin, dmax, interval):
        if not interval:
            if dmax == dmin:
                interval = 10
            else:
                interval = 10 ** (float(int(math.log(dmax - dmin) / math.log(10))))
        dmin = min(dmin, pychart_util.round_down(dmin, interval))
        dmax = max(dmax, pychart_util.round_up(dmax, interval) + interval / 2.0)
        return dmin, dmax, interval



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from six import string_types, integer_types
import types

AnyType = 9998


def IntervalType(val):
    if type(val) in (int, int,
                     float, types.FunctionType):
        return None
    return "Expecting a number or a function"


def CoordType(val):
    if type(val) not in (tuple, list):
        return "Expecting a tuple or a list"
    if len(val) != 2:
        return "Coordinate must be a pair of numbers"
    for v in val:
        if v != None and NumberType(v):
            return "Expecting a pair of numbers (got %s)" % str(v)
    return None


def NumberType(val):
    if type(val) in (int, int, float):
        return None
    else:
        return "Expecting a number"


def UnitType(val):
    if type(val) in (int, int, float):
        return
    else:
        return "Expecting a unit"


def ShadowType(val):
    if type(val) not in (tuple, list):
        return "Expecting tuple or list."
    if len(val) != 3:
        return "Expecting (xoff, yoff, fill)."
    return None


def FormatType(val):
    if type(val) in (bytes, types.FunctionType):
        return None
    return "Format must be a string or a function"


def StringType(val):
    if type(val) in string_types:
        return None
    return "Expected a string"


def IntType(val):
    if type(val) in integer_types:
        return None
    return "Expected a integer"


def ListType(val):
    if type(val) is list:
        return None
    return "Expected a list"


def TupleType(val):
    if type(val) is tuple:
        return None
    return "Expected a tuple"



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
#
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import os
import re

from . import area
from . import arrow
from . import axis
from . import bar_plot
from . import color
from . import doc_support
from . import error_bar
from . import fill_style
from . import legend
from . import line_plot
from . import line_style
from . import pie_plot
from . import range_plot
from . import text_box
from . import tick_mark

indent = 4
max_line_len = 64


def format_paragraph(fp, str):
    line_len = indent
    fp.write(" " * indent)
    for word in str.split():
        if line_len >= max_line_len:
            fp.write("\n")
            fp.write(" " * indent)
            line_len = indent
        fp.write(word + " ")
        line_len += len(word) + 1


def format_string(fp, str):
    str = re.sub("<<([^>]+)>>", "See also pychart.\\1", str)

    str2 = ""
    in_example = 0
    for l in str.split("\n"):
        if re.match("@example", l):
            in_example = 1
        if re.match("@end example", l):
            in_example = 0
        if in_example:
            str2 += l
        else:
            l = re.sub("^[ \t]*", "", l)
            str2 += l
        str2 += "\n"
    fname = os.tempnam()
    out_fp = open(fname, "w")
    out_fp.write(str2)
    out_fp.close()

    in_fp = os.popen("makeinfo --fill-column=64 --no-headers " + fname, "r")
    for l in in_fp.readlines():
        fp.write(" " * indent)
        fp.write(l)
    in_fp.close()
    os.remove(fname)


def generate_doc(c, name, suffix="", append=0):
    if append:
        fp = open(name + "_doc.py", "a+")
    else:
        fp = open(name + "_doc.py", "w")
        fp.write("# automatically generated by generate_docs.py.\n")

    fp.write("doc" + suffix + "=\"\"\"Attributes supported by this class are:\n")
    for key in list(c.keys.keys()):
        val = c.keys[key]
        desc = ""
        defaultValDesc = None
        if len(val) > 3:
            desc = val[3]
        if len(val) > 4:
            defaultValDesc = val[4]

        fp.write(key + "(type:" + doc_support.stringify_type(val[0]))
        if defaultValDesc:
            fp.write(") default:" + defaultValDesc)
        else:
            fp.write(") default=" + str(doc_support.stringify_value(val[2])) + ".\n")
        format_string(fp, desc)
    fp.write("\"\"\"\n\n")
    fp.close()


generate_doc(arrow.T, "arrow")
generate_doc(area.T, "area")
generate_doc(axis.X, "axis", "_x")
generate_doc(axis.Y, "axis", "_y", 1)
generate_doc(bar_plot.T, "bar_plot")
generate_doc(line_plot.T, "line_plot")
generate_doc(pie_plot.T, "pie_plot")
generate_doc(color.T, "color")
generate_doc(error_bar.error_bar1, "error_bar", "_1")
generate_doc(error_bar.error_bar2, "error_bar", "_2", 1)
generate_doc(error_bar.error_bar3, "error_bar", "_3", 1)
generate_doc(error_bar.error_bar4, "error_bar", "_4", 1)
generate_doc(error_bar.error_bar5, "error_bar", "_5", 1)
generate_doc(error_bar.error_bar6, "error_bar", "_6", 1)
generate_doc(fill_style.T, "fill_style")
generate_doc(text_box.T, "text_box")
generate_doc(range_plot.T, "range_plot")
generate_doc(legend.T, "legend")
generate_doc(legend.Entry, "legend", "_entry", 1)
generate_doc(line_style.T, "line_style")
generate_doc(tick_mark.T, "tick_mark")
